import json

from future.utils import iteritems

from tcell_agent.appsensor.params import GET_PARAM, POST_PARAM, JSON_PARAM, COOKIE_PARAM, URI_PARAM
from tcell_agent.config import CONFIGURATION
from tcell_agent.sensor_events import AppSensorEvent

# set is builtin in python 3 but a module in earlier version
try:
    set
except NameError:
    from sets import Set as set

import logging
PAYLOADS_LOGGER = logging.getLogger('app_firewall_payloads_logger').getChild(__name__)

def get_payloads_logger():
    return PAYLOADS_LOGGER

class PayloadsPolicy(object):

    DEFAULT_BLACKLIST = {
        "token": set(['*']),
        "client_secret": set(['*']),
        "password": set(['*']),
        "passwd": set(['*']),
        "refresh_token": set(['*']),
        "pf.pass": set(['*']),
        "user.password": set(['*']) }

    PARAM_TYPE_MAP = {
        GET_PARAM: "form",
        POST_PARAM: "form",
        JSON_PARAM: "form",
        URI_PARAM: "form",
        COOKIE_PARAM: "cookie" }


    def __init__(self, policy_json=None):
        self.reset()

    def reset(self):
        self.send_payloads = False
        self.send_payloads = False
        self.log_payloads = False

        self.send_blacklist = self.DEFAULT_BLACKLIST
        self.log_blacklist = self.DEFAULT_BLACKLIST
        self.send_whitelist = {}
        self.log_whitelist = {}

        self.use_send_whitelist = False
        self.use_log_whitelist = False

    def apply(self, dp, appsensor_meta, type_of_param, vuln_param, vuln_value, meta, pattern):
        payload = None

        if self.send_payloads and CONFIGURATION.allow_unencrypted_appsensor_payloads:

            blacklisted_locations = self.send_blacklist.get(vuln_param.lower(), None)
            param_location = self.PARAM_TYPE_MAP[type_of_param]

            if blacklisted_locations and ( param_location in blacklisted_locations or "*" in blacklisted_locations ):
                payload = "BLACKLISTED"

            elif self.use_send_whitelist:
                whitelisted_locations = self.send_whitelist.get(vuln_param.lower(), None)
                if whitelisted_locations and ( param_location in whitelisted_locations or "*" in whitelisted_locations ):
                  payload = vuln_value

                else:
                  payload = "NOT_WHITELISTED"

            else:
              payload = vuln_value


        self.log(dp, appsensor_meta, type_of_param, vuln_param, vuln_value, meta, pattern)

        return payload

    def log(self, dp, appsensor_meta, type_of_param, vuln_param, vuln_value, meta, pattern):
        if self.log_payloads and CONFIGURATION.log_appfirewall_events:
            blacklisted_locations = self.log_blacklist.get(vuln_param.lower(), None)
            param_location = self.PARAM_TYPE_MAP[type_of_param]

            if blacklisted_locations is None or \
                ( param_location not in blacklisted_locations and "*" not in blacklisted_locations ):

                whitelisted_locations = self.log_whitelist.get(vuln_param.lower(), None)
                if not self.use_log_whitelist or \
                    ( whitelisted_locations is not None and \
                    ( param_location in whitelisted_locations or "*" in whitelisted_locations ) ):

                    event = AppSensorEvent(dp,
                        vuln_param,
                        appsensor_meta.location,
                        appsensor_meta.remote_address,
                        appsensor_meta.route_id,
                        meta,
                        appsensor_meta.method,
                        payload=vuln_value,
                        user_id=appsensor_meta.user_id,
                        hmacd_session_id=appsensor_meta.session_id,
                        pattern=pattern)

                    get_payloads_logger().info(json.dumps(event))

    def from_json(self, policy_json):
        self.reset()

        if policy_json:
            payloads_json = policy_json.get("payloads", {})
            self.send_payloads = payloads_json.get("send_payloads", False)
            self.log_payloads = payloads_json.get("log_payloads", False)

            if self.send_payloads:
                send_blacklist = payloads_json.get("send_blacklist", None)
                if send_blacklist is not None:
                    self.send_blacklist = {}
                    for param_name, locations in iteritems(payloads_json.get("send_blacklist", {})):
                        self.send_blacklist[param_name.lower()] = set(locations)

                send_whitelist = payloads_json.get("send_whitelist", None)
                if send_whitelist is not None:
                    for param_name, locations in iteritems(send_whitelist):
                        self.send_whitelist[param_name.lower()] = set(locations)

                    self.use_send_whitelist = True

            if self.log_payloads:
                log_blacklist = payloads_json.get("log_blacklist", None)
                if log_blacklist is not None:
                    self.log_blacklist = {}
                    for param_name, locations in iteritems(payloads_json.get("log_blacklist", {})):
                        self.log_blacklist[param_name.lower()] = set(locations)

                log_whitelist = payloads_json.get("log_whitelist", None)
                if log_whitelist is not None:
                    for param_name, locations in iteritems(log_whitelist):
                        self.log_whitelist[param_name.lower()] = set(locations)
                    self.use_log_whitelist = True



from __future__ import unicode_literals
from __future__ import print_function
import unittest
from ...config import CONFIGURATION
from ...agent import TCellAgent, PolicyTypes
import json

class AgentMetricsMetricsCountTest(unittest.TestCase):
    def setUp(self):
        CONFIGURATION.version = 1
        CONFIGURATION.api_key = "Test_-ApiKey=="
        CONFIGURATION.app_id = "TestAppId-AppId"
        CONFIGURATION.host_identifier = "TestHostIdentifier"
        CONFIGURATION.uuid = "test-uuid-test-uuid"
        CONFIGURATION.fetch_policies_from_tcell = False

        added_events = []
        self.added_events = added_events
        def addEvent(self, event):
          event.post_process()
          added_events.append(event)

        @classmethod
        def send(cls, event):
          cls.get_agent().addEvent(event)

        self.old_add_events = TCellAgent.addEvent
        TCellAgent.addEvent = addEvent
        self.old_send = TCellAgent.send
        TCellAgent.send = send

        self.old_tcell_agent = TCellAgent.tCell_agent
        TCellAgent.tCell_agent = TCellAgent()
        TCellAgent.tCell_agent.policies = {}

    def tearDown(self):
        TCellAgent.addEvent = self.old_add_events
        TCellAgent.send = self.old_send
        TCellAgent.tCell_agent = self.old_tcell_agent

    def test_event_on_metrics_accumulated(self):
        policy_one = """
        {
          "login":{
            "policy_id":"00a1",
            "data": {
              "options": {
                 "session_hijacking_enabled":true,
                 "login_success_enabled":true
              }
            }
          }
        }
        """
        policy_json = json.loads(policy_one)
        TCellAgent.get_agent().process_policies(policy_json, cache=False)

        login_policy = TCellAgent.get_policy(PolicyTypes.LOGIN)

        old_send = TCellAgent.send
        @classmethod
        def send(cls, event):
          cls.get_agent().addEvent(event)
          print(event)
          if (event["event_type"] == "dummy"):
              cls.get_agent().agent_metrics.get_and_reset_sesssion_metric()
        TCellAgent.send = send
        try:
            for sessionid_num in range(0,300):
                session_id = "a"*40 + str(sessionid_num)
                TCellAgent.request_metric("route_1",300,"123.333.444.555","UserAgent x 4/05.31.2/2@23",session_id=session_id,user_id="bob@bob.com")

            self.assertEqual(self.added_events[0]["event_type"], "dummy")
            self.assertEqual(len(self.added_events), 3)
            self.assertEqual(TCellAgent.get_agent().agent_metrics.get_object_count(), 144)
        finally:
            TCellAgent.send = old_send

    def test_event_on_metrics_accumulated_policy_off(self):
        policy_one = """
        {
         "login":{
          "policy_id":"00a1",
          "data": {
            "options": {
               "session_hijacking_enabled":false,
               "login_success_enabled":true
            }
          }
         }
        }
        """
        policy_json = json.loads(policy_one)
        TCellAgent.get_agent().process_policies(policy_json, cache=False)

        login_policy = TCellAgent.get_policy(PolicyTypes.LOGIN)

        old_send = TCellAgent.send
        @classmethod
        def send(cls, event):
          cls.get_agent().addEvent(event)
          if (event["event_type"] == "dummy"):
              cls.get_agent().agent_metrics.get_and_reset_sesssion_metric()
        TCellAgent.send = send
        try:
            for sessionid_num in range(0,300):
                session_id = "a"*40 + str(sessionid_num)
                TCellAgent.request_metric("route_1",300,"123.333.444.555","UserAgent x 4/05.31.2/2@23",session_id=session_id,user_id="bob@bob.com")

            self.assertEqual(len(self.added_events), 0)
            self.assertEqual(TCellAgent.get_agent().agent_metrics.get_object_count(), 0)
        finally:
            TCellAgent.send = old_send

    def test_count_of_metrics(self):
        from ...agent.metrics import AgentMetrics

        metrics = AgentMetrics()
        metrics.add_session_track_metric("sessid1", "userid1", "useragent2", "1.2.2.2")
        self.assertEqual(metrics.get_object_count(), 3)
        metrics.get_and_reset_sesssion_metric()
        self.assertEqual(metrics.get_object_count(), 0)



    def test_event_on_route_metrics_accumulated(self):
        old_send = TCellAgent.send
        @classmethod
        def send(cls, event):
          cls.get_agent().addEvent(event)
          cls.get_agent().agent_metrics.get_and_reset_route_count_table()
          print(TCellAgent.get_agent().agent_metrics._route_count_table)
          print(event)
        TCellAgent.send = send
        try:
            for route in range(0,320):
                route_id = "a"*40 + str(route)
                TCellAgent.request_metric(route_id,300,"123.333.444.555","UserAgent x 4/05.31.2/2@23","abcdefg",user_id="bob@bob.com")
            self.assertEqual(len(self.added_events), 2)
            self.assertEqual(self.added_events[0]["event_type"], "dummy")
            self.assertEqual(len(TCellAgent.get_agent().agent_metrics._route_count_table), 20)
        finally:
            TCellAgent.send = old_send        



import re


class AppSensorRulePattern(object):
    def __init__(self, pattern_id, pattern_regex, enabled=True):
        self.pattern_id = pattern_id
        self.pattern_regex = pattern_regex
        self.enabled = enabled

class AppSensorRuleSet(object):
    def __init__(self):
        self.safe_pattern = None
        self.patterns = []

    def check_violation(self, param_name, param_value, active_pattern_ids, v1_compatability_enabled):
        try:
            if param_value is None or (self.safe_pattern and self.safe_pattern.search(param_value)):
                return None
            for pattern in self.patterns:
                if pattern is None or pattern.enabled == False:
                    continue
                if v1_compatability_enabled or active_pattern_ids.get(pattern.pattern_id, False):
                    match = pattern.pattern_regex.search(param_value)
                    if match:
                        return {"m": match, "param": param_name, "value": param_value, "pattern": pattern.pattern_id}
        except:
            pass
        return None

    def read_from_rules_file(filename):
        appsensor_ruleset = AppSensorRuleSet()
        #appsensor_ruleset.safe_pattern = rules

    def add_pattern_from_dict(self, rule_dict):
        if rule_dict is None:
            return

        pattern_id = rule_dict.get("id")
        pattern = rule_dict.get("python", None)
        if pattern is None:
            pattern = rule_dict.get("common", None)
        elif pattern == "disabled":
            return

        if pattern_id is None or pattern is None:
            return

        pattern_regex = re.compile(pattern, flags=re.IGNORECASE|re.S|re.M)
        enabled = rule_dict.get("enabled", True)

        rule_pattern = AppSensorRulePattern(pattern_id, pattern_regex, enabled)
        self.add_pattern(rule_pattern)

    def set_safe_pattern_from_string(self, safe_pattern_str):
        if (safe_pattern_str is not None):
            self.safe_pattern = re.compile(safe_pattern_str, flags=re.IGNORECASE|re.S|re.M)

    def add_pattern(self, pattern):
        self.patterns.append(pattern)



from __future__ import unicode_literals
from __future__ import print_function

import re
import unittest

from tcell_agent.policies.appsensor.payloads_policy import PayloadsPolicy

# set is builtin in python 3 but a module in earlier version
try:
    set
except NameError:
    from sets import Set as set

class PayloadsPolicyTest(unittest.TestCase):

    def from_json_none_options_test(self):
        policy_json = None
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_none_payloads_test(self):
        policy_json = {}
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_empty_payloads_test(self):
        policy_json = { "payloads": {} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_false_send_payloads_test(self):
        policy_json = { "payloads": {"send_payloads": False} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_false_send_payloads_with_send_blacklist_send_whitelist_test(self):
        policy_json = {
            "payloads": {
                "send_payloads": False,
                "send_blacklist": {
                    "username": ["*"]
                },
                "send_whitelist": {
                    "password": ["*"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_true_send_payloads_test(self):
        policy_json = { "payloads": {"send_payloads": True} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertTrue(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_true_send_payloads_with_send_blacklist_send_whitelist_test(self):
        policy_json = {
            "payloads": {
                "send_payloads": True,
                "send_blacklist": {
                    "username": ["*"],
                    "password": ["form"]
                },
                "send_whitelist": {
                    "password": ["*"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertTrue(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, {"username": set(["*"]), "password": set(["form"])})
        self.assertEqual(policy.send_whitelist, {"password": set(["*"])})
        self.assertTrue(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_false_log_payloads_test(self):
        policy_json = { "payloads": {"log_payloads": False} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_false_log_payloads_with_log_blacklist_log_whitelist_test(self):
        policy_json = {
            "payloads": {
                "log_payloads": False,
                "log_blacklist": {
                    "username": ["*"]
                },
                "log_whitelist": {
                    "password": ["*"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertFalse(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_true_log_payloads_test(self):
        policy_json = { "payloads": {"log_payloads": True} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertTrue(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.log_whitelist, {})
        self.assertFalse(policy.use_log_whitelist)

    def from_json_payloads_true_log_payloads_with_log_blacklist_log_whitelist_test(self):
        policy_json = {
            "payloads": {
                "log_payloads": True,
                "log_blacklist": {
                    "username": ["*"],
                    "password": ["form"]
                },
                "log_whitelist": {
                    "password": ["*"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)

        self.assertFalse(policy.send_payloads)
        self.assertTrue(policy.log_payloads)
        self.assertEqual(policy.send_blacklist, policy.DEFAULT_BLACKLIST)
        self.assertEqual(policy.send_whitelist, {})
        self.assertFalse(policy.use_send_whitelist)
        self.assertEqual(policy.log_blacklist, {"username": set(["*"]), "password": set(["form"])})
        self.assertEqual(policy.log_whitelist, {"password": set(["*"])})
        self.assertTrue(policy.use_log_whitelist)



from tcell_agent.policies.appsensor import InjectionSensor


class XssSensor(InjectionSensor):

    def __init__(self, policy_json=None):
        super(XssSensor, self).__init__("xss", policy_json)

        self.libinjection = False

        if policy_json is not None:
            self.libinjection = policy_json.get("libinjection", False)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" API Module handles all calls to/from the tCell service. Errors are
handled gracefully since it's generally running silently and should
fail open.
"""

from __future__ import unicode_literals
from __future__ import print_function

import threading
import requests
import json
import logging
from tcell_agent.config import CONFIGURATION
from tcell_agent.version import VERSION
import sys

LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

class SetEncoder(json.JSONEncoder):    
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)

    def encode(self, obj):
        def traverse(obj):
            if isinstance(obj, dict):
                return {k: traverse(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [traverse(elem) for elem in obj]
            elif (isinstance(obj, str) or isinstance(obj, bytes) or (sys.version_info < (3, 0) and isinstance(obj, unicode))) and len(obj) > 256:
                return obj[:256]
            else:
                return obj  # no container, just values (str, int, float)
        return super(SetEncoder, self).encode(traverse(obj))


class TCellAPIException(Exception):
    """Special API Exception"""
    pass

class TCellAPI(object):
    """API Class handling comms"""

    def v1update(self, last_timestamp=None):
        """v1 is the new rest API endpoint"""
        url = '{url}/app/{appname}/update'.format(
            url=CONFIGURATION.tcell_api_url,
            appname=CONFIGURATION.app_id)
        LOGGER.debug("calling api: " + url)
        params = {}
        if last_timestamp:
            params["last_timestamp"] = last_timestamp
        headers = {"Authorization":"Bearer " + CONFIGURATION.api_key,
                   "TCellAgent":"Python " + VERSION}
        try:
            response = requests.get(url, params=params, headers=headers, allow_redirects=False)
        except Exception as general_exception:
            LOGGER.debug(general_exception)
            raise TCellAPIException("could not connect to server")
        LOGGER.debug("response: " + str(response))
        if response.ok:
            try:
                response_json = response.json()
                if response_json.get("result") is not None:
                    result = response_json.get("result")
                    return result
                else:
                    return None
            except Exception as general_exception:
                LOGGER.debug(general_exception)
                raise TCellAPIException("Result field not found")
        LOGGER.debug(response)
        raise TCellAPIException("Response was not 'ok'")

    def v1send_events(self, events, event_wrap_name=None, event_endpoint=None):
        """
        Sends events to tCell via the API
        v1 is the new rest API
        """
        if event_wrap_name is None:
            event_wrap_name = "events"
        if event_endpoint is None:
            event_endpoint = "server_agent"
        url = '{url}/app/{appname}/{endpoint}'.format(
            url=CONFIGURATION.tcell_input_url,
            appname=CONFIGURATION.app_id,
            endpoint=event_endpoint)
        payload = {"hostname": CONFIGURATION.host_identifier,
                   "uuid": CONFIGURATION.uuid
                  }
        payload[event_wrap_name] = events
        LOGGER.debug("sending events to " + url)
        LOGGER.debug(json.dumps(payload, cls=SetEncoder))
        headers = {"Authorization":"Bearer " + CONFIGURATION.api_key,
                   "Content-type":"application/json",
                   "TCellAgent":"Python " + VERSION}
        response = requests.post(url, data=json.dumps(payload, cls=SetEncoder), headers=headers, allow_redirects=False)
        LOGGER.debug("send_events response: " + str(response))
        # who cares if it succeeded?
        return response

    def v0api(self, api_call, last_timestamp=None):
        """Some legacy support for old configs, don't use this"""
        url = '{url}/api/{appname}{api_call}/{secret}'.format(
            url=CONFIGURATION.tcell_api_url,
            appname=CONFIGURATION.app_id,
            secret=CONFIGURATION.api_key,
            api_call=api_call)
        params = {}
        if last_timestamp:
            params["last_timestamp"] = last_timestamp
        LOGGER.debug("calling OLD api: " + url)
        response = requests.get(url, params=params)
        LOGGER.debug("old API response: " + str(response))
        if response.ok:
            try:
                response_json = response.json()
                if response_json.get("result") is not None:
                    result = response_json.get("result")
                    return result
                else:
                    return None
            except Exception as general_exception:
                LOGGER.exception(general_exception)
                raise TCellAPIException("Result field not found")
        raise TCellAPIException("Response Failed")

    @classmethod
    def send_events(cls, events, event_wrap_name=None, event_endpoint=None):
        """Want to manually send events? This is your method!"""
        try:
            cls.static_api
        except AttributeError:
            cls.static_api = cls()
        try:
            return cls.static_api.v1send_events(
                events,
                event_wrap_name=event_wrap_name,
                event_endpoint=event_endpoint)
        except Exception as sending_exception:
            LOGGER.error("Exception while sending TCell event.")
            LOGGER.exception(sending_exception)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from future.standard_library import install_aliases

#install_aliases()
#from urllib.parse import urlparse, urlencode

from tcell_agent.policies import TCellPolicy

import logging
logger = logging.getLogger('tcell_agent').getChild(__name__)

from future.backports.urllib.parse import urlparse
from future.backports.urllib.parse import urlencode
from future.backports.urllib.parse import urlsplit
from future.backports.urllib.parse import urlunsplit
from future.backports.urllib.parse import parse_qsl
from future.backports.urllib.parse import urlunparse
from future.backports.urllib.parse import quote_plus

def java_hash(s):
    h = 0
    for c in s:
        h = (31 * h + ord(c)) & 0xFFFFFFFF
    return ((h + 0x80000000) & 0xFFFFFFFF) - 0x80000000
    
class ContentSecurityHeader:
    def __init__(self, csp_type, value, report_uri=None, policy_id=None):
        self.csp_type = None
        self.value = None
        self.policy_id = None
        if (csp_type is not None):
            if (csp_type.lower() == "content-security-policy"):
                self.csp_type = "csp"
            elif (csp_type.lower() == "content-security-policy-report-only"):
                self.csp_type = "csp-report"
            elif (csp_type == "csp" or csp_type == "csp-report"):
                self.csp_type = csp_type
            else:
                raise Exception("Unknown header")
        self.report_uri = report_uri
        self.value = value
        self.policy_id = policy_id

    def header_value(self, transaction_id=None, route_id=None, session_id=None, user_id=None):
        if self.report_uri is None:
            return self.value
        try:
            query_parts = []
            if transaction_id:
                query_parts.append("tid=" + quote_plus(transaction_id))
            if session_id:
                query_parts.append("sid=" + quote_plus(session_id))
            if route_id:
                query_parts.append("rid=" + quote_plus(route_id))

            if len(query_parts) > 0:
                report_url = self.report_uri + "?" + "&".join(query_parts)
            else:
                report_url = self.report_uri

            if self.policy_id is not None:
                checksum = str(java_hash(self.policy_id + report_url))
                if len(query_parts) > 0:
                    report_url = report_url + "&"
                else:
                    report_url = report_url + "?"                    
                report_url = report_url + "c=" + checksum
            return "%s; report-uri %s" % (self.value, report_url)
        except Exception as e:
            logger.debug("error creating csp: %s: (%s, %s), falling back" % (e, self.value, self.report_uri))
            return "%s; report-uri %s" % (self.value, self.report_uri)

    def header_names(self):
        if (self.csp_type == "csp"):
            return ["Content-Security-Policy"]
        elif (self.csp_type == "csp-report"):
            return ["Content-Security-Policy-Report-Only"]
        else:
            return []

class ContentSecurityPolicy(TCellPolicy):
    def __init__(self, policy_json=None, header_class=ContentSecurityHeader):
        super(ContentSecurityPolicy, self).__init__()
        self.header_class = header_class
        self.csp_headers = []
        self.js_agent_api_key = None
        if policy_json is not None:
            self.loadFromJson(policy_json)

    def loadFromHeaderString(self, header_string):
        if header_string is not None:
            csp_header_split = header_string.split(":",1)
            csp_headers = []
            try:
                csp_header = self.header_class(csp_header_split[0], csp_header_split[1])
                csp_headers.append(csp_header)
            except:
                logger.error("error parsing secure header (%s)" % (header_string))
            self.csp_headers = csp_headers

    def headers(self, transaction_id=None, route_id=None, session_id=None, user_id=None):
        headers = []
        for header in self.csp_headers:
            header_value = header.header_value(transaction_id, route_id, session_id, user_id)
            for header_name in header.header_names():
                headers.append([header_name, header_value])
        return headers

    def loadFromJson(self, policy_json):
        if "policy_id" in policy_json:
            self.policy_id = policy_json["policy_id"]
        else:
            raise Exception("Policy Id Not Found")
        csp_headers = []
        self.js_agent_api_key = None
        if "data" in policy_json:
            if "options" in policy_json["data"]:
                policy_option_json = policy_json["data"]["options"]
                if policy_option_json:
                    self.js_agent_api_key = policy_option_json.get("js_agent_api_key", None)

        if "headers" in policy_json:
            for header in policy_json["headers"]:
                if "name" in header and "value" in header:
                    try:
                        csp_header = self.header_class(header["name"], header["value"], header.get("report-uri"), self.policy_id)
                        csp_headers.append(csp_header)
                    except Exception as e:
                        logger.error("error parsing csp header (%s, %s)" % (header["name"], header["value"]))
                pass
        

        self.csp_headers = csp_headers
        




# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent
from tcell_agent.sanitize import SanitizeUtils
from future.backports.urllib.parse import urlsplit

class DiscoveryEvent(SensorEvent):
    DATABASE_TYPE = "db"

    def __init__(self,
                 route_id=None):
        super(DiscoveryEvent, self).__init__("discovery")
        if route_id is not None:
            self["rid"] = route_id

    def for_database(self, database, schema, table, field):
        self["type"] = "db"
        self["db"] = database
        self["schema"] = schema
        self["table"] = table
        self["field"] = field
        return self

    def for_database_fields(self, database, schema, table, fields):
        self["type"] = "db"
        self["db"] = database
        self["schema"] = schema
        self["table"] = table
        self["fields"] = fields
        return self


# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent
from tcell_agent.sanitize import SanitizeUtils
from future.backports.urllib.parse import urlsplit

class HoneytokenSensorEvent(SensorEvent):
    def __init__(self,
        remote_addr,
        token_id
        ):
        super(HoneytokenSensorEvent, self).__init__("honeytoken")
        self["id"] = token_id
        self["remote_addr"] = remote_addr

    def post_process(self):
        pass


from tcell_agent.policies.appsensor import SizeSensor

class ResponseSizeSensor(SizeSensor):
    MAX_NORMAL_RESPONSE_BYTES = 1024*1024*2
    DP_UNUSUAL_RESPONSE_SIZE = "rspsz"

    def __init__(self, policy_json=None):
        super(ResponseSizeSensor, self).__init__(
            self.MAX_NORMAL_RESPONSE_BYTES,
            self.DP_UNUSUAL_RESPONSE_SIZE,
            policy_json
        )



from tcell_agent.policies.appsensor import InjectionSensor

class CmdiSensor(InjectionSensor):

    def __init__(self, policy_json=None):
        super(CmdiSensor, self).__init__("cmdi", policy_json)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

import json
import os.path
import socket
import uuid
import os
import errno
import yaml
import logging
import logging.handlers
import traceback

DEFAULT_CONFIG_FILENAME = "tcell_agent.config"
DEFAULT_HOME = "tcell/"

def mkdir_p(path):
    """http://stackoverflow.com/a/600612/190597 (tzot)"""
    try:
        os.makedirs(path, exist_ok=True)  # Python>3.2
    except TypeError:
        try:
            os.makedirs(path)
        except OSError as exc: # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else: raise

class TcellAgentConfiguration(object):
    """docstring for ClassName"""
    def get_configuration_logger(self, log_filename):
        new_logger = logging.getLogger("tcell_agent_configuration")
        formatter = logging.Formatter('%(asctime)s - %(name)s -- [%(levelname)s]: %(message)s')

        fh = logging.handlers.RotatingFileHandler(
                  log_filename, maxBytes=10*1024*1024, backupCount=5)
        fh.setFormatter(formatter)

        new_logger.setLevel("INFO")
        new_logger.addHandler(fh)
        return new_logger

    def __init__(self, filename=DEFAULT_CONFIG_FILENAME):
        self.tcell_agent_home = './tcell_agent'
        self.tcell_agent_log_home = None
        self.tcell_agent_config = None

        if os.environ.get('TCELL_AGENT_ALLOW_UNENCRYPTED_APPSENSOR_PAYLOADS', None) != None:
            print("tCell.io Agent: [DEPRECATED] TCELL_AGENT_ALLOW_UNENCRYPTED_APPSENSOR_PAYLOADS is deprecated, please switch to TCELL_AGENT_ALLOW_UNENCRYPTED_APPFIREWALL_PAYLOADS.")

        self.allow_unencrypted_appsensor_payloads = \
            (os.environ.get('TCELL_AGENT_ALLOW_UNENCRYPTED_APPSENSOR_PAYLOADS') in [True, "true", "1"]) or \
            (os.environ.get('TCELL_AGENT_ALLOW_UNENCRYPTED_APPFIREWALL_PAYLOADS') in [True, "true", "1"])



        self.max_data_ex_db_records_per_request = 1000

        self.version = 0
        self.app_id = None
        self.api_key = None
        self.tcell_api_url = "https://api.tcell.io/api/v1"
        self.tcell_input_url = "https://input.tcell.io/api/v1"
        self.company = None
        self.uuid = str(uuid.uuid4())
        self.host_identifier = None
        self.preload_policy_filename = None
        self.fetch_policies_from_tcell = True
        self.instrument_for_events = True
        self.hmac_key = None
        self.logging_options = {}
        self.reverse_proxy = True
        self.reverse_proxy_ip_address_header = None

        try:
            self.demo_mode = bool(os.environ.get('TCELL_DEMOMODE', None))
        except:
            self.demo_mode = False

        try:
            self.event_batch_size_limit = int(os.environ.get('TCELL_EVENT_SIZE_BATCH_LIMIT', 100))
        except:
            self.event_batch_size_limit = 100
        self.event_time_limit_seconds = 15

        self.event_queue_intake_size = 1000

        self.js_agent_api_base_url = None
        self.js_agent_url = "https://api.tcell.io/tcellagent.min.js"

        self.tcell_home = os.environ.get("TCELL_AGENT_HOME", DEFAULT_HOME)
        self.cache_folder = os.path.join(self.tcell_home, 'cache/')
        self.log_directory = os.environ.get("TCELL_AGENT_LOG_DIR", os.path.join(self.tcell_home, "logs/"))
        self.log_filename = os.path.join(self.log_directory, "tcell_agent.log")

        mkdir_p(self.tcell_home)
        mkdir_p(self.cache_folder)
        mkdir_p(self.log_directory)

        config_filename = os.path.abspath(os.environ.get("TCELL_AGENT_CONFIG",filename))
        backup_check_config_filename = os.path.abspath(os.path.join("tcell",filename))

        config_logger = self.get_configuration_logger(self.log_filename)

        config_logger.info("Checking for configuration file: '" + config_filename + "'")

        try:
            if os.path.isfile(config_filename) == False:
                config_logger.info("Checking for configuration file: '" + backup_check_config_filename + "'")
                if os.path.isfile(backup_check_config_filename) == True:
                    config_filename = backup_check_config_filename
                else:
                    config_filename = None

            if config_filename != None:
                if os.access(config_filename, os.R_OK) == False:
                    print("tCell.io Agent: [Error] Permission denied opening config file.")
                    config_logger.error("Permission denied for file '" + config_filename + "'")
                else:
                    with open(config_filename, "r") as data_file:
                        json_data = self.filter_commentlines(data_file.read())
                        try:
                            config_json = json.loads(json_data)
                            self.read_config(config_json)
                        except:
                            print("tCell.io Agent: [Error] Could not parse json in config file.")
                            traceback.print_exc()
                            config_logger.exception("Could not parse json in config file.")
            else:
                config_logger.info("Configuration file not found, using environmental variables.")
        except Exception as ex:
            print("Exception loading configuration file.")
            print(ex)
            config_logger.exception("Exception loading configuration file.")
            raise

        self.log_appfirewall_events = os.environ.get('TCELL_AGENT_ALLOW_UNENCRYPTED_APPFIREWALL_PAYLOADS_LOGGING') in [True, "true", "1"]
        self.app_firewall_payloads_log_filename = os.path.join(self.log_directory, "tcell_agent_payloads.log")

        config_logger.info("Checking configuration environmental variables.")

        self.app_id = os.environ.get("TCELL_AGENT_APP_ID", self.app_id)
        self.api_key = os.environ.get("TCELL_AGENT_API_KEY", self.api_key)
        self.host_identifier = os.environ.get("TCELL_AGENT_HOST_IDENTIFIER", self.host_identifier)

        if self.app_id is None:
            config_logger.warn("tCell.io APP ID not set by file or environmental variable.")
        if self.api_key is None:
            config_logger.warn("tCell.io API Key not set by file or environmental variable.")

    def filter_commentlines(self, json_data):
        """
        Removes full line comments from json to provide trivial support for commenting. Does not remove
        end of line comments.
        :param json_data: A json text string
        :return: json text without comments
        """
        return "\n".join(l for l in json_data.split("\n") if not l.strip().startswith("//"))

    def read_config(self, config_json):
        if "version" not in config_json:
            self.version = 0
            self.company = config_json.get("COMPANY")
            self.app_id = config_json.get("APP_NAME")
            self.api_key = config_json.get("API_KEY")
            self.tcell_input_url = config_json.get("TCELL_INPUT_URL","https://input.tcell.io:8008")
            self.host_identifier = socket.getfqdn()
        else:
            self.version = config_json["version"]
            if self.version == 1:
                applications = config_json["applications"]
                if applications and len(applications) > 0:
                    application_config_json = applications[0] # Get the first application for now
                    v0_app_id = application_config_json.get("name")
                    self.app_id = application_config_json.get("app_id", v0_app_id)
                    self.api_key = application_config_json.get("api_key")
                    self.enabled = application_config_json.get("enabled",True)
                    self.fetch_policies_from_tcell = application_config_json.get("fetch_policies_from_tcell", True)
                    self.preload_policy_filename = application_config_json.get("preload_policy_filename")
                    self.logging_options = application_config_json.get("logging_options", self.logging_options)
                    self.instrument_for_events = application_config_json.get("instrument_for_events",True)
                    new_api_url = application_config_json.get("tcell_api_url",  None)
                    if (new_api_url is not None and new_api_url != self.tcell_api_url):
                        # Change default jsagent
                        self.js_agent_api_base_url = new_api_url
                        self.tcell_api_url = new_api_url

                    self.tcell_input_url = application_config_json.get("tcell_input_url", self.tcell_input_url)
                    host_identifier = application_config_json.get("host_identifier")
                    if host_identifier is None:
                        host_identifier = socket.getfqdn()
                    self.host_identifier = host_identifier

                    self.hmac_key = application_config_json.get("hmac_key")

                    self.js_agent_api_base_url = application_config_json.get("js_agent_api_base_url",self.js_agent_api_base_url)
                    self.js_agent_url = application_config_json.get("js_agent_url",self.js_agent_url)

                    # hidden
                    try:
                        self.event_batch_size_limit = int(application_config_json.get('event_batch_size_limit', self.event_batch_size_limit))
                    except:
                        pass

                    self.allow_unencrypted_appsensor_payloads = application_config_json.get('allow_unencrypted_appsensor_payloads', self.allow_unencrypted_appsensor_payloads) 
                    self.allow_unencrypted_appsensor_payloads = application_config_json.get('allow_unencrypted_appfirewall_payloads', self.allow_unencrypted_appsensor_payloads) 

                    (os.environ.get('TCELL_AGENT_ALLOW_APPSENSOR_PAYLOADS', '0')=='1')

                    data_exposure = application_config_json.get('data_exposure', {})
                    self.max_data_ex_db_records_per_request = data_exposure.get('max_data_ex_db_records_per_request', self.max_data_ex_db_records_per_request)

                    self.reverse_proxy = application_config_json.get('reverse_proxy', self.reverse_proxy)
                    self.reverse_proxy_ip_address_header = application_config_json.get('reverse_proxy_ip_address_header', self.reverse_proxy_ip_address_header)

                    self.demo_mode = application_config_json.get('demomode', self.demo_mode)
                    if self.demo_mode:
                        self.event_batch_size_limit = 1
                        self.event_time_limit_seconds = 5

                    #for old tcell_input_url
                    self.company = application_config_json.get("company")



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.appsensor import params

class AppSensorMeta(object):
    def __init__(
            self,
            remote_address,
            method,
            location,
            route_id,
            session_id=None,
            user_id=None,
            transaction_id=None):
        self.remote_address = remote_address
        self.method = method
        self.location = location
        self.route_id = route_id
        self.session_id = session_id
        self.user_id = user_id
        self.transaction_id = transaction_id

        self.get_dict = None
        self.post_dict = None
        self.cookie_dict = None
        self.files_dict = None
        self.json_body_str = None
        self.request_content_len = None
        self.response_code = None
        self.do_request = False
        self.do_response = False
        self.user_agent_str = False

        self.do_path_parameters = False
        self.path_dict = None

    def request_data(
            self,
            request_content_len,
            get_dict,
            post_dict,
            cookie_dict,
            json_body_str,
            user_agent_str,
            files_dict):
        self.do_request = True
        self.get_dict = get_dict
        self.cookie_dict = cookie_dict
        self.json_body_str = json_body_str
        self.request_content_len = request_content_len
        self.user_agent_str = user_agent_str

        filenames_dict = {}
        for param_name in (files_dict or {}).keys():
            filenames_dict[param_name] = []
            for file_obj in files_dict.getlist(param_name):
                filenames_dict[param_name].append(file_obj.name)

        self.files_dict = params.flatten_clean(filenames_dict)
        self.post_dict = params.flatten_clean(post_dict)

    def response_data(self, response_content_len=None, response_code=None):
        self.do_response = True
        self.response_content_len = response_content_len
        self.response_code = response_code

    def path_parameters_data(self, path_dict):
        self.do_path_parameters = True
        self.path_dict = path_dict




from __future__ import unicode_literals
from __future__ import print_function

import unittest

from mock import patch
from tcell_agent.appsensor.meta import AppSensorMeta
from tcell_agent.policies.appsensor import ResponseCodesSensor

class ResponseCodesSensorTest(unittest.TestCase):

    def create_default_sensor_test(self):
        sensor = ResponseCodesSensor()
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.series_400_enabled, False)
        self.assertEqual(sensor.series_500_enabled, False)

    def create_enabled_sensor_test(self):
        sensor = ResponseCodesSensor({"enabled": True})
        self.assertEqual(sensor.enabled, True)
        self.assertEqual(sensor.series_400_enabled, False)
        self.assertEqual(sensor.series_500_enabled, False)

    def create_sensor_with_series_400_enabled_test(self):
        sensor = ResponseCodesSensor({"series_400_enabled": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.series_400_enabled, True)
        self.assertEqual(sensor.series_500_enabled, False)

    def create_sensor_with_series_500_enabled_test(self):
        sensor = ResponseCodesSensor({"series_500_enabled": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.series_400_enabled, False)
        self.assertEqual(sensor.series_500_enabled, True)

    def with_disabled_sensor_check_test(self):
        sensor = ResponseCodesSensor({"enabled": False})
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check({}, 200)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_200_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 200)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_300_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 300)
            self.assertFalse(patched_send_event.called)

    def with_disabled_400_series_and_400_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": False, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 400)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_400_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 400)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[4], None, {"code": 400})

    def with_enabled_sensor_and_400_response_code_and_matching_excluded_route_id_check_test(self):
        sensor = ResponseCodesSensor({
            "enabled": True,
            "series_400_enabled": True,
            "series_500_enabled": True,
            "exclude_routes": ["23947"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 400)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_400_response_code_and_nonmatching_excluded_route_id_check_test(self):
        sensor = ResponseCodesSensor({
            "enabled": True,
            "series_400_enabled": True,
            "series_500_enabled": True,
            "exclude_routes": ["nonmatching"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 400)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[4], None, {"code": 400})

    def with_enabled_sensor_and_401_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 401)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[401], None, {"code": 401})

    def with_enabled_sensor_and_403_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 403)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[403], None, {"code": 403})

    def with_enabled_sensor_and_404_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 404)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[404], None, {"code": 404})

    def with_disabled_500_series_and_500_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": False})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 500)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_500_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 500)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[500], None, {"code": 500})

    def with_enabled_sensor_and_501_response_code_check_test(self):
        sensor = ResponseCodesSensor({"enabled": True, "series_400_enabled": True, "series_500_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.response_codes_sensor.sendEvent') as patched_send_event:
            sensor.check(meta, 501)
            patched_send_event.assert_called_once_with(meta, ResponseCodesSensor.RESPONSE_CODE_DP_DICT[5], None, {"code": 501})



from .size_sensor import SizeSensor
from .request_size_sensor import RequestSizeSensor
from .response_size_sensor import ResponseSizeSensor
from .response_codes_sensor import ResponseCodesSensor

from .injection_sensor import InjectionSensor
from .xss_sensor import XssSensor
from .sqli_sensor import SqliSensor
from .cmdi_sensor import CmdiSensor
from .fpt_sensor import FptSensor
from .nullbyte_sensor import NullbyteSensor
from .retr_sensor import RetrSensor
from .login_sensor import LoginSensor

from .user_agent_sensor import UserAgentSensor
from .misc_sensor import MiscSensor
from .database_sensor import DatabaseSensor

from .payloads_policy import PayloadsPolicy



from tcell_agent.policies.appsensor.sensor import sendEvent

class SizeSensor(object):
    MAX_NORMAL_REQUEST_BYTES = 1024*512
    DP_UNUSUAL_REQUEST_SIZE = "reqsz"

    def __init__(self, default_limit, dp_code, policy_json=None):
        self.enabled = False
        self.limit = default_limit
        self.dp_code = dp_code
        self.excluded_route_ids = {}

        if policy_json is not None:
            self.enabled = policy_json.get("enabled", False)
            self.limit = policy_json.get("limit", self.limit)

            for route_id in policy_json.get("exclude_routes", []):
                self.excluded_route_ids[route_id] = True

    def check(self, appsensor_meta, str_length):
        if not self.enabled or self.excluded_route_ids.get(appsensor_meta.route_id, False):
            return

        if (str_length > self.limit):
            sendEvent(
              appsensor_meta,
              self.dp_code,
              None,
              { "sz": str_length})

    def __str__(self):
        return "<%s enabled: %s limit: %s dp_code: %s excluded_route_ids: %s>" % \
            (type(self).__name__, self.enabled, self.limit, self.dp_code, self.excluded_route_ids)



from __future__ import unicode_literals
from __future__ import print_function

import unittest
import json
from ...policies.dataloss_policy import DataLossPolicy


policy_one = """
{
  "v":1,
  "policy_id":"xyz-def",
  "data":{
    "db_protections":[
      {"tables":["work_infos"],
       "fields":["SSN"],
       "actions":{"body":["redact"]}
      },
      {"tables":["work_infos"],
       "fields":["income"],
       "actions":{"body":["redact"], "log":["redact"]}
      }
    ]
  }
}
"""

policy_db_protection_by_route = """
{
  "v":1,
  "policy_id":"xyz-def",
  "data":{
    "db_protections":[
      {"scope":"route",
       "route_ids":["abcdefg"],
       "tables":["work_infos"],
       "fields":["SSN"],
       "actions":{"body":["redact"]}
      },
      {"tables":["work_infos"],
       "fields":["income"],
       "actions":{"body":["redact"], "log":["redact"]}
      }
    ]
  }
}
"""

policy_session_protection_one = """
{
  "v":1,
  "policy_id":"xyz-def",
  "data":{
    "session_id_protections":{"body":["redact"], "log":["event"]}
  }
}
"""

policy_data_discovery = """
{
  "v":1,
  "policy_id":"xyz-def",
  "data":{
    "data_discovery":{
      "database_enabled":true
    },
    "session_id_protections":{"body":["redact"], "log":["event"]}
  }
}"""



class DataLossPolicyTest(unittest.TestCase):

    def session_test(self):
        policy_json = json.loads(policy_session_protection_one)
        policy = DataLossPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.get_actions_for_session_id().body_redact, True)
        self.assertEqual(policy.get_actions_for_session_id().log_event, True)


    def normal_policy_test(self):
        policy_json = json.loads(policy_one)
        policy = DataLossPolicy()
        policy.loadFromJson(policy_json)
        actions_ssn = policy.get_actions_for_db_field("*","*","work_infos","SSN")
        action_ssn = list(actions_ssn)[0]
        self.assertEqual(action_ssn.body_redact, True)
        self.assertEqual(action_ssn.log_redact, False)
        actions_income = policy.get_actions_for_db_field("test","test","work_infos","income")
        action_income = list(actions_income)[0]
        self.assertEqual(action_income.body_redact, True)
        self.assertEqual(action_income.log_redact, True)

    def normal_policy_test_using_route(self):
        policy_json = json.loads(policy_db_protection_by_route)
        policy = DataLossPolicy()
        policy.loadFromJson(policy_json)

        actions_ssn = policy.get_actions_for_db_field("*","*","work_infos","SSN",route="abcdefg")
        action_ssn = list(actions_ssn)[0]
        self.assertEqual(action_ssn.body_redact, True)
        self.assertEqual(action_ssn.log_redact, False)

        actions_income = policy.get_actions_for_db_field("test","test","work_infos","income")
        action_income = list(actions_income)[0]
        self.assertEqual(action_income.body_redact, True)
        self.assertEqual(action_income.log_redact, True)

    def discovery_policy_test(self):
        policy_json = json.loads(policy_data_discovery)
        policy = DataLossPolicy()
        self.assertEqual(policy.database_discovery_enabled, False)
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.database_discovery_enabled, True)

        #self.assertEqual(policy.get_actions_for_django('"work_infos"."income"'), ["body_redact","log_redact"])
        #self.assertEqual(policy.get_actions_for_django('"work_infos"."SSN"'), ["body_redact"])



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals

from future.utils import iteritems

GET_PARAM = "get"
POST_PARAM = "post"
JSON_PARAM = "json"
COOKIE_PARAM = "cookies"
URI_PARAM = "uri"

# source: http://stackoverflow.com/questions/38987/how-can-i-merge-two-python-dictionaries-in-a-single-expression
def merge_two_dicts(x, y):
    '''Given two dicts, merge them into a new dict as a shallow copy.'''
    z = x.copy()
    z.update(y)
    return z

def flatten_clean(param_dict, namespace=None):
    result = {}
    if namespace is None:
        namespace = tuple()

    for param_name, param_value in iteritems(param_dict or {}):
        if isinstance(param_value, dict):
            new_namespace = namespace + (str(param_name),)
            result = merge_two_dicts(result, flatten_clean(param_value, new_namespace))

        elif isinstance(param_value, list):
            for idx, val in enumerate(param_value):
                new_namespace = namespace + (str(idx), str(param_name))

                if isinstance(val, dict):
                    result = merge_two_dicts(result, flatten_clean(val, new_namespace))

                else:
                    # this is for python3
                    if isinstance(val, bytes):
                        val = val.decode('utf-8')

                    param_type = str(type(val))

                    # have to do it this way because unicode class does not exist in python3
                    # so isinstance check can't be used
                    if param_type.startswith("<class 'str'") or \
                            param_type.startswith("<type 'str'") or \
                            param_type.startswith("<type 'unicode'"):
                        result[new_namespace] = val


        else:
            # this is for python3
            if isinstance(param_value, bytes):
                param_value = param_value.decode('utf-8')

            param_type = str(type(param_value))

            # have to do it this way because unicode class does not exist in python3
            # so isinstance check can't be used
            if param_type.startswith("<class 'str'") or \
                    param_type.startswith("<type 'str'") or \
                    param_type.startswith("<type 'unicode'"):
                new_key = namespace + (param_name,)
                result[new_key] = param_value

    return result



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""

from __future__ import unicode_literals
from __future__ import print_function
import tcell_agent
from .system_info import mkdir_p
import json
import logging
import os
import threading

LOGGER = logging.getLogger('tcell_agent').getChild(__name__)
cacheFileLock = threading.Lock()

class TCellPolicyCache(object):
    """Manages cache."""
    def __init__(self, app_id, cache_dir='tcell/cache'):
        self.app_id = app_id
        mkdir_p(cache_dir)
        self.cache_filename = cache_dir + '/tcell_app.' + self.app_id
        self.pid = os.getpid()
        self.master_cache = {}

    def get_lock(self):
        return cacheFileLock
        
    def _read_cache(self):
        cache_json = {}
        try:
            if os.path.isfile(self.cache_filename) == True:
                with open(self.cache_filename) as data_file:
                    json_data = data_file.read()
                    cache_json = json.loads(json_data)
                data_file.close()
        except Exception as e:
            LOGGER.debug("Error loading cache file")
            LOGGER.debug(e)
            return self.master_cache
        self.master_cache.update(cache_json)
        return self.master_cache

    def current_cache(self):
        self.get_lock().acquire()
        try:
            self._read_cache()
        finally:
            self.get_lock().release()
        return self.master_cache

    def update_cache(self, policy_name, policy_json):
        try:
            self.get_lock().acquire()
            self.master_cache.update({policy_name: policy_json})
            try:
                with open(self.cache_filename, "w") as data_file:
                    json.dump(self.master_cache, data_file, indent=2)
                data_file.close()
            finally:
                self.get_lock().release()        
        except Exception as e:
            LOGGER.debug("Error writing cache file")
            LOGGER.debug(e)
        return            



from __future__ import unicode_literals
from __future__ import print_function
import unittest
from ...sensor_events import ServerAgentPackagesEvent
#from ... import tcell_agent

class SanitizerUtilsTest(unittest.TestCase):
    def test_package_event_create(self):
        sape = ServerAgentPackagesEvent()
        sape.add_package("test_package", "test_version")
        self.assertEqual(sape["packages"], [{"n":"test_package", "v":"test_version"}])

    def test_packages_event_create(self):
        sape = ServerAgentPackagesEvent()
        sape.add_package("test_package", "test_version")
        sape.add_package("test_package2", "test_version2")
        self.assertEqual(sape["packages"], [
          {"n":"test_package", "v":"test_version"},
          {"n":"test_package2", "v":"test_version2"}
          ])


# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function

from django.test import TestCase
from django.test.client import RequestFactory
from django.http import HttpResponse
import time

from future.backports.urllib.parse import urlparse, urlencode
from future.backports.urllib.parse import urlsplit
from future.backports.urllib.parse import urlunsplit
from future.backports.urllib.parse import parse_qs


try:
    import django
    from django.conf import settings
    settings.configure()
    settings.DATABASES = {
    'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'mydatabase',
    }}
    settings.ROOT_URLCONF = 'tcell_agent.tests.instrumentation.django.middleware.test_urls'
    django.setup()
except RuntimeError:
    print("Django already setup")

from mock import Mock
from .....policies.appsensor_policy import AppSensorPolicy

from .....instrumentation.django.middleware.globalrequestmiddleware import GlobalRequestMiddleware
from .....instrumentation.django.middleware.afterauthmiddleware import AfterAuthMiddleware
from .....instrumentation.django.middleware.body_filter_middleware import BodyFilterMiddleware
from .....instrumentation.django.middleware.tcell_data_exposure_middleware import TCellDataExposureMiddleware
from .....instrumentation.django.middleware.tcelllastmiddleware import TCellLastMiddleware

from django.middleware.common import CommonMiddleware

from .....config import CONFIGURATION
from .....agent import TCellAgent, PolicyTypes


from django.conf import settings
try:
    settings.configure()
except:
    pass
settings.DEFAULT_CHARSET = 'utf-8'

class AppSensorMiddlewareTest(TestCase):

    def setUp(self):
        settings.ALLOWED_HOSTS = ['testserver','test.tcell.io']

        self.grm = GlobalRequestMiddleware()
        self.aam = AfterAuthMiddleware()
        self.cm = CommonMiddleware()
        self.bfm = BodyFilterMiddleware()
        self.data_exposure_middleware = TCellDataExposureMiddleware()
        self.tcell_last_middleware = TCellLastMiddleware()

        self.rf = RequestFactory(HTTP_USER_AGENT="Mozilla/5.0")

        CONFIGURATION.version = 1
        CONFIGURATION.api_key = "Test_-ApiKey=="
        CONFIGURATION.app_id = "TestAppId-AppId"
        CONFIGURATION.host_identifier = "TestHostIdentifier"
        CONFIGURATION.uuid = "test-uuid-test-uuid"
        CONFIGURATION.fetch_policies_from_tcell = False

        added_events = []
        self.added_events = added_events
        def addEvent(self, event):
          event.post_process()
          added_events.append(event)

        self.old_add_events = TCellAgent.addEvent
        TCellAgent.addEvent = addEvent

        self.old_tcell_agent = TCellAgent.tCell_agent
        TCellAgent.tCell_agent = TCellAgent()
        TCellAgent.tCell_agent.policies = {}

    def request_with_session(self,
        url='http://test.tcell.io/hello/',
        session_key="101012301200123",
        content_type=None,
        post_form_data=None,
        host=None):
        if post_form_data is not None:
            if content_type is not None:
                request = self.rf.post(url, content_type=content_type, data=post_form_data)
            else:
                request = self.rf.post(url, data=post_form_data)
        else:
            request = self.rf.get(url)
        request.session = Mock()
        request.session.session_key = session_key
        return request

    def middleware_before(self, request, route_id="mock_request_id"):
        self.grm.process_request(request)
        request._tcell_context.route_id = route_id
        self.data_exposure_middleware.process_request(request)
        self.aam.process_request(request)
        self.cm.process_request(request)
        self.bfm.process_request(request)
        self.tcell_last_middleware.process_request(request)

    def middleware_after(self, request, response):
        self.tcell_last_middleware.process_response(request, response)
        self.bfm.process_response(request, response)
        self.cm.process_response(request, response)
        self.aam.process_response(request, response)
        self.data_exposure_middleware.process_response(request, response)
        self.grm.process_response(request, response)

    def test_appsensor_xss(self):
        request = self.request_with_session(url='http://test.tcell.io/hello/?someparam=<SCRIPT>alert(1)</SCRIPT>')
        request.META["SERVER_NAME"] = "test.tcell.io"
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd",
                "data": {
                    "options": {
                        "xss":True,
                        "sqli":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 1)
        appsensor_event = self.added_events[0]

        self.assertEqual(appsensor_event["dp"],"xss")
        self.assertEqual(appsensor_event["meta"],{"l": "query"})
        self.assertEqual(appsensor_event["param"],"someparam")
        self.assertEqual(appsensor_event["m"],"GET")
        self.assertEqual(appsensor_event["uri"],"http://test.tcell.io/hello/?someparam=")
        self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")

    def test_appsensor_sqli_and_xss(self):
        request = self.request_with_session(url='http://test.tcell.io/hello/?someparam=<script>alert(1)</script>&def=%30"%20OR%20"2"="2')
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd", 
                "data": {
                    "options": {
                        "xss":True,
                        "sqli":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)
        old_allow = CONFIGURATION.allow_unencrypted_appsensor_payloads
        CONFIGURATION.allow_unencrypted_appsensor_payloads = False
        try:
            self.middleware_before(request)
            response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
            self.middleware_after(request, response)

            self.assertEqual(len(self.added_events), 2)
            sorted_events = sorted(self.added_events, key=lambda item:item["dp"])
            appsensor_event = sorted_events[1]

            self.assertEqual(appsensor_event["dp"],"xss")
            self.assertEqual(appsensor_event["meta"],{"l": "query"})
            self.assertEqual(appsensor_event["param"],"someparam")
            self.assertEqual(appsensor_event["m"],"GET")
            self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")
            self.assertEqual(appsensor_event.get("payload"),None)
            self.assertEqual(("payload" in appsensor_event),False)

            appsensor_event = sorted_events[0]

            self.assertEqual(appsensor_event["dp"],"sqli")
            self.assertEqual(appsensor_event["param"],"def")
            self.assertEqual(appsensor_event["m"],"GET")
            self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")
            self.assertEqual(appsensor_event.get("payload"),None)
            self.assertEqual(("payload" in appsensor_event),False)
        finally:
            CONFIGURATION.allow_unencrypted_appsensor_payloads = old_allow

    def test_appsensor_post(self):
        request = self.request_with_session(
            url='http://test.tcell.io/hello/',
            post_form_data={'name':'fred', 'passwd':'secret<SCRIPT>prompt("xss")</SCRIPT>'})
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd", 
                "data": {
                    "options": {
                        "xss":True,
                        "sqli":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 1)
        appsensor_event = self.added_events[0]

        self.assertEqual(appsensor_event["dp"],"xss")
        self.assertEqual(appsensor_event["meta"],{"l": "body"})
        self.assertEqual(appsensor_event["param"],"passwd")
        self.assertEqual(appsensor_event["m"],"POST")
        self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")

    def test_appsensor_json_xss(self):
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd", 
                "data": {
                    "options": {
                        "xss":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        body = """{"a":[{"b":"<script>alert(1)</script>"}]}"""
        request = self.request_with_session(post_form_data=body, content_type="application/json")
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 1)
        appsensor_event = self.added_events[0]

        self.assertEqual(appsensor_event["dp"],"xss")
        self.assertEqual(appsensor_event["meta"],{"l": "body"})
        self.assertEqual(appsensor_event["param"],"b")
        self.assertEqual(appsensor_event["m"],"POST")
        self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")

    def test_appsensor_json_xss_type_form(self):
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd", 
                "data": {
                    "options": {
                        "xss":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        body = """{"a=":[{"b":"<script>alert(1)</script>"}]}"""
        request = self.request_with_session(post_form_data=body, content_type="application/x-www-form-urlencoded")
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 1)
        appsensor_event = self.added_events[0]

        self.assertEqual(appsensor_event["dp"],"xss")
        self.assertEqual(appsensor_event["meta"],{"l": "body"})
        self.assertEqual(appsensor_event["param"],"b")
        self.assertEqual(appsensor_event["m"],"POST")
        self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")

    def test_appsensor_invalid_json_xss_type_form(self):
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd", 
                "data": {
                    "options": {
                        "xss":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        body = """{"a="[{"b":"<script>alert(1)</script>"}]}"""
        request = self.request_with_session(post_form_data=body, content_type="application/x-www-form-urlencoded")
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 1)
        appsensor_event = self.added_events[0]

        self.assertEqual(appsensor_event["dp"],"xss")
        self.assertEqual(appsensor_event["meta"],{"l": "body"})
        self.assertEqual(appsensor_event["param"],"{\"a")
        self.assertEqual(appsensor_event["m"],"POST")
        self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")

    def test_appsensor_json_clean(self):
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd",
                "data": {
                    "options": {
                        "xss":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        body = """{"a":[{"b":"this is not an xss script"}]}"""
        request = self.request_with_session(post_form_data=body, content_type="application/json")
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 0)

    def test_appsensor_request_size(self):
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd",
                "data": {
                    "options": {
                        "req_res_size":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        request = self.request_with_session(url='http://test.tcell.io/hello/', post_form_data={"A":"B"})
        request.META["CONTENT_LENGTH"] = str((1024*512)+1)
        request.META["SERVER_NAME"] = "test.tcell.io"
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 1)
        appsensor_event = self.added_events[0]
        self.assertEqual(appsensor_event["dp"],"reqsz")
        self.assertEqual(appsensor_event["uri"],"http://test.tcell.io/hello/")

    def test_appsensor_request_size_small(self):
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd", 
                "data": {
                    "options": {
                        "req_res_size":True,
                        "cmdi":True
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        request = self.request_with_session(post_form_data={"A":"B"})
        request.META["CONTENT_LENGTH"] = (1024*512)-1
        self.middleware_before(request)
        response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
        self.middleware_after(request, response)

        self.assertEqual(len(self.added_events), 0)

    def test_appsensor_payloads(self):
        request = self.request_with_session(url='http://test.tcell.io/hello/abc?someparam=<script>alert(1)</script>&def=%30"%20OR%20"2"="2')
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd",
                "data": {
                    "options": {
                        "xss":True,
                        "sqli":True,
                        "cmdi":True
                    }
                }
            }
        }

        old_allow = CONFIGURATION.allow_unencrypted_appsensor_payloads
        CONFIGURATION.allow_unencrypted_appsensor_payloads = True
        try:
            TCellAgent.tCell_agent.process_policies(policy_json, cache=False)
            self.middleware_before(request)
            response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
            self.middleware_after(request, response)
            self.assertEqual(len(self.added_events), 2)
            sorted_events = sorted(self.added_events, key=lambda item:item["dp"])
            appsensor_event = sorted_events[1]

            self.assertEqual(appsensor_event["dp"],"xss")
            self.assertEqual(appsensor_event["meta"],{"l": "query"})
            self.assertEqual(appsensor_event["param"],"someparam")
            self.assertEqual(appsensor_event["m"],"GET")
            self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")

            self.assertEqual(appsensor_event["payload"],"<script>alert(1)</script>")

            #some sort of ordering randomness
            #self.assertEqual(appsensor_event["uri"],"/hello/abc?someparam=&def=")
            scheme, netloc, path, query_string, fragment = urlsplit(appsensor_event["uri"])
            query_params = parse_qs("query_string")
            self.assertEqual(len(query_params), 0)

            appsensor_event = sorted_events[0]

            self.assertEqual(appsensor_event["dp"],"sqli")
            self.assertEqual(appsensor_event["param"],"def")
            self.assertEqual(appsensor_event["m"],"GET")
            self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")
            self.assertEqual(appsensor_event["payload"],'0" OR "2"="2')
        finally:
            CONFIGURATION.allow_unencrypted_appsensor_payloads = old_allow

    def test_appsensor_payloads_default_blacklist(self):
        request = self.request_with_session(url='http://test.tcell.io/hello/abc?password=<script>alert(1)</script>&def=%30"%20OR%20"2"="2')
        policy_json = {
            "appsensor": {
                "policy_id":"nyzd",
                "data": {
                    "options": {
                        "xss":True,
                        "sqli":True,
                        "cmdi":True
                    }
                }
            }
        }

        old_allow = CONFIGURATION.allow_unencrypted_appsensor_payloads
        CONFIGURATION.allow_unencrypted_appsensor_payloads = True
        try:
            TCellAgent.tCell_agent.process_policies(policy_json, cache=False)
            self.middleware_before(request)
            response =  HttpResponse("<html>\n<head>Title</head><body>hello world: 101012301200123 101012301200123</body><html>",content_type="application/html")
            self.middleware_after(request, response)

            self.assertEqual(len(self.added_events), 2)
            sorted_events = sorted(self.added_events, key=lambda item:item["dp"])
            appsensor_event = sorted_events[1]

            self.assertEqual(appsensor_event["dp"],"xss")
            self.assertEqual(appsensor_event["meta"],{"l": "query"})
            self.assertEqual(appsensor_event["param"],"password")
            self.assertEqual(appsensor_event["m"],"GET")
            self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")
            self.assertEqual(appsensor_event["payload"],"BLACKLISTED")

            #some sort of ordering randomness
            #self.assertEqual(appsensor_event["uri"],"/hello/abc?someparam=&def=")
            scheme, netloc, path, query_string, fragment = urlsplit(appsensor_event["uri"])
            query_params = parse_qs("query_string")
            self.assertEqual(len(query_params), 0)

            appsensor_event = sorted_events[0]

            self.assertEqual(appsensor_event["dp"],"sqli")
            self.assertEqual(appsensor_event["param"],"def")
            self.assertEqual(appsensor_event["m"],"GET")
            self.assertEqual(appsensor_event["remote_addr"],"127.0.0.1")
            self.assertEqual(appsensor_event["payload"],'0" OR "2"="2')
        finally:
            CONFIGURATION.allow_unencrypted_appsensor_payloads = old_allow



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""

from __future__ import unicode_literals
from __future__ import print_function
from collections import defaultdict
from ..sensor_events import DiscoveryEvent
from json import JSONEncoder

def inner_dict(next):
  return lambda: defaultdict(next)


class DataExfilRule(object):
    def __init__(self):
        self.body_event = False
        self.body_redact = False
        self.body_hash = False
        self.log_event = False
        self.log_redact = False
        self.log_hash = False
        self.id = 0

class DatabaseFieldEndpoint(dict):
    def __init__(self):
        self.discovered = False
        self.data_exfil_rule = None

class RouteEndpoint(object):
    def __init__(self):
        self.discovered = False
        # self.database_fields = inner_dict(inner_dict(inner_dict(inner_dict(DatabaseFieldEndpoint))))()
        self.database_fields = inner_dict(DatabaseFieldEndpoint)()
        self.database_exfil_rules = inner_dict(inner_dict(inner_dict(inner_dict(DatabaseFieldEndpoint))))()

    def hash_database_tuple(self, database, schema, table, fields):
        return hash(database + "," + schema + "," + table + "," + ",".join(fields))

    def has_field_been_discovered(self, database,schema,table,field):
        return self.have_fields_been_discovered(database, schema, table, [field])

    def have_fields_been_discovered(self, database,schema,table,fields):
        if (database and schema and table and fields):
            if (self.database_fields[self.hash_database_tuple(database,schema,table,fields)].discovered == False):
                return True
        return False

    def set_field_discovered(self, database, schema, table, field):
        self.set_fields_discovered(database, schema, table, [field])

    def set_fields_discovered(self, database, schema, table, fields):
        if (database and schema and table and fields):
            self.database_fields[self.hash_database_tuple(database,schema,table,fields)].discovered = True


class RouteTable(object):
    def __init__(self):
        self.routes = defaultdict(RouteEndpoint)
        self.update_cache_flag = False        
        #self.full_database = DiscoveryEndpoint()




# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function


from tcell_agent.policies import TCellPolicy
from tcell_agent.sensor_events import DlpEvent
from collections import defaultdict
import collections

"""
  "dlp": {
    "v":1,
    "policy_id":"xyz-def",
    "data":{
      "protections":[
        {"table":"work_infos",
         "field":"SSN",
         "actions":{"body":["redact"], "log":["redact"]}
        },
        {"table":"work_infos",
         "field":"income",
         "actions":{"body":["redact"], "log":["redact"]}
        }
      ]
    }
  },
"""
"""
    "request_protections": [
      {
        "id": "129321",
        "scope": "route",
        "route_ids": ["234234", "34432432"],
        "variable_context": "form",
        "variables": ["password"],
        "actions": {
          "body": ["event"],
          "log": ["redact"]
        }
      },
"""

class DataLossPolicyEnforcement(object):
    def __init__(self):
        self.body_event = False
        self.body_redact = False
        self.body_hash = False
        self.log_event = False
        self.log_redact = False
        self.log_hash = False
        self.value_regex = None
        self.id = 0
    def apply_regex(self, value):
        # TODO: apply the regex to the value
        return value

class RequestProtectionManager(object):
    FORM = "form"
    COOKIE = "cookie"
    HEADER = "header"
    def __init__(self):
        def inner_dict(next):
          return lambda: collections.defaultdict(next)
        # ["route_id"]["parameter"] = set(Rule1, Rule2)
        self.parameter_protections = inner_dict(inner_dict(set))()
        self.cookie_protections = inner_dict(inner_dict(set))()
        self.header_protections = inner_dict(inner_dict(set))()
    def add_form_protection(self, form_parameter, rule, route_id="*"):
        if form_parameter is not None:
            self.parameter_protections[form_parameter.lower()][route_id].add(rule)
    def get_rules_for_form_parameter(self, form_parameter, route_id="*"):
        return_actions = set()
        if form_parameter in  self.parameter_protections:
            if (route_id in self.parameter_protections[form_parameter]):
                return_actions = return_actions.union(self.parameter_protections[form_parameter][route_id])
            elif (route_id != "*" and "*" in self.parameter_protections[form_parameter]):
                return_actions = return_actions.union(self.parameter_protections[form_parameter]["*"])
        if len(return_actions) == 0:
            return None
        return return_actions
    def add_cookie_protection(self, cookie_name, rule, route_id="*"):
        if cookie_name:
            self.cookie_protections[cookie_name][route_id].add(rule)
    def get_rules_for_cookie(self, cookie_name, route_id="*"):
        return_actions = set()
        if cookie_name in  self.cookie_protections:
            if (route_id in self.cookie_protections[cookie_name]):
                return_actions = return_actions.union(self.cookie_protections[cookie_name][route_id])
            elif (route_id != "*" and "*" in self.cookie_protections[cookie_name]):
                return_actions = return_actions.union(self.cookie_protections[cookie_name]["*"])
        if len(return_actions) == 0:
            return None
        return return_actions

    def add_header_protection(self, header_name, rule, route_id="*"):
        if header_name:
            self.header_protections[header_name][route_id].add(rule)
    def get_rules_for_header(self, header_name, route_id="*"):
        return_actions = set()
        if header_name in  self.header_protections:
            if (route_id in self.header_protections[header_name]):
                return_actions = return_actions.union(self.header_protections[header_name][route_id])
            elif (route_id != "*" and "*" in self.header_protections[header_name]):
                return_actions = return_actions.union(self.header_protections[header_name]["*"])
        if len(return_actions) == 0:
            return None
        return return_actions


class DataLossPolicy(TCellPolicy):
    REDACT_MESSAGE = "[redact]"
    B_REDACT_MESSAGE = b"[redact]"
    api_identifier = "dlp"

    def __init__(self, policy_json=None):
        super(DataLossPolicy, self).__init__()
        self.init_fields()
        if policy_json is not None:
            self.loadFromJson(policy_json)

    def init_fields(self):
        self.database_discovery_enabled = False

        self.field_redact_body = set()
        self.field_alerts = set()

        self.session_id_actions = []
        self.table_field_actions = defaultdict(list)

        def inner_dict(next):
          return lambda: collections.defaultdict(next)

        # [db][schema][table][field] = actions
        self.database_protections = inner_dict(inner_dict(inner_dict(inner_dict(inner_dict(set)))))()
        self.request_protections = RequestProtectionManager()
        self.session_id_filter_actions = None

    def get_actions_for_session_id(self, route_id=None):
        return self.session_id_filter_actions 

    def get_actions_for_db_field(self, database, schema, table, field, route="*"):
        return_actions = set()
        for d in (database,"*"):
            if d not in self.database_protections:
                continue
            for s in (schema, "*"):
                if s not in self.database_protections[d]:
                    continue
                for t in (table, "*"):
                    if t not in self.database_protections[d][s]:
                        continue
                    for f in (field, "*"):
                        if f in self.database_protections[d][s][t]:
                            if (route in self.database_protections[d][s][t][f]):
                                return_actions = return_actions.union(self.database_protections[d][s][t][f][route])
                            elif (route != "*" and "*" in self.database_protections[d][s][t][f]):
                                return_actions = return_actions.union(self.database_protections[d][s][t][f]["*"])
        if len(return_actions)==0:
            return None
        return return_actions

    def get_actions_for_form_parameter(self, parameter, route_id="*"):
        return self.request_protections.get_rules_for_form_parameter(parameter,route_id)

    def get_actions_for_cookie(self, cookie_name, route_id="*"):
        return self.request_protections.get_rules_for_cookie(cookie_name,route_id)

    def get_actions_for_header(self, header_name, route_id="*"):
        return self.request_protections.get_rules_for_header(header_name,route_id)

    def loadFromJson(self, policy_json):
        self.init_fields()
        if "policy_id" in policy_json:
            self.policy_id = policy_json["policy_id"]
        else:
            raise Exception("Policy Id Not Found")

        policy_data_json = policy_json.get("data")

        def actions_from_json(protections):
            actions = None
            body = protections.get("body", [])
            if "redact" in body:
                actions = actions or DataLossPolicyEnforcement()
                actions.body_redact = True
            if "event" in body:
                actions = actions or DataLossPolicyEnforcement()
                actions.body_event = True
            if "hash" in body:
                actions = actions or DataLossPolicyEnforcement()
                actions.body_hash = True
            log = protections.get("log", [])
            if "redact" in log:
                actions = actions or DataLossPolicyEnforcement()
                actions.log_redact = True
            if "event" in log:
                actions = actions or DataLossPolicyEnforcement()
                actions.log_event = True
            if "hash" in log:
                actions = actions or DataLossPolicyEnforcement()
                actions.log_hash = True
            return actions

        if policy_data_json:
            data_discovery_json = policy_data_json.get("data_discovery")
            if data_discovery_json is not None:
                self.database_discovery_enabled = data_discovery_json.get("database_enabled", False)
            session_id_protection = policy_data_json.get("session_id_protections",[])
            if session_id_protection:
                self.session_id_filter_actions = actions_from_json(session_id_protection)
            request_protections = policy_data_json.get("request_protections",[])
            for protection in request_protections:
                """
                        "id": "129321",
                        "scope": "route",
                        "route_ids": ["234234", "34432432"],
                        "variable_context": "form",
                        "variables": ["password"],
                """
                scope = protection.get("scope")
                variable_context = protection.get("variable_context",None)
                variables = protection.get("variables",None)
                actions = protection.get("actions",{})
                rule_id = protection.get("id",0)
                if variable_context is None or variables is None:
                    continue
                route_ids = []
                if (scope is None or scope == "global"):
                    route_ids = ["*"]
                elif (scope == "route"):
                    route_ids = protection.get("route_ids",[])
                rule = actions_from_json(actions)
                if rule is None:
                    continue
                rule.id = rule_id
                for route_id in route_ids:
                    for variable in variables:
                        if variable_context == RequestProtectionManager.FORM or variable_context == "parameter":
                            self.request_protections.add_form_protection(variable.lower(), rule, route_id)
                        elif variable_context == RequestProtectionManager.COOKIE:
                            #cookie names are case sensitive
                            self.request_protections.add_cookie_protection(variable, rule, route_id)
                        elif variable_context == RequestProtectionManager.HEADER:
                            self.request_protections.add_header_protection(variable.lower(), rule, route_id)                            

            protections = policy_data_json.get("db_protections",[])
            for protection in protections:
                databases = protection.get("databases",["*"])
                scope = protection.get("scope")
                schemas = protection.get("schemas",["*"])
                tables = protection.get("tables",["*"])
                fields = protection.get("fields",None)
                actions = protection.get("actions",{})
                rule_id = protection.get("id",0)
                
                if fields is None:
                    continue
                route_ids = []
                if (scope is None or scope == "global"):
                    route_ids = ["*"]
                if (scope == "route"):
                    route_ids = protection.get("route_ids",[])
                field_actions = actions_from_json(actions)
                if field_actions is None:
                    continue
                field_actions.id = rule_id
                if (databases and schemas and tables and fields):
                    if ("*" in databases):
                        databases = ["*"]
                    for database in databases:
                        if ("*" in schemas):
                            schemas = ["*"]
                        for schema in schemas:
                            if ("*" in tables):
                                tables = ["*"]
                            for table in tables:
                                if ("*" in fields):
                                    fields = ["*"]
                                for field in fields:
                                    if ("*" in route_ids):
                                        route_ids = ["*"]
                                    for route_id in route_ids:
                                        self.database_protections[database][schema][table][field][route_id].add(field_actions)
                # self.database_protections = []
                # key = '"{table}"."{field}"'.format(table=table, field=field)
                # body_actions = actions.get("body",[])
                # if "redact" in body_actions:
                #     self.table_field_actions[key].append("body_redact")
                # log_actions = actions.get("log",[])
                # if "redact" in log_actions:
                #     self.table_field_actions[key].append("log_redact")
            # print("*********************************** MM")
            # for d in self.database_protections:
            #     for s in self.database_protections[d]:
            #         for t in self.database_protections[d][s]:
            #             for f in self.database_protections[d][s][t]:
            #                 print(d + "," + s + "," + t + "," + f)
            #                 print(self.database_protections[d][s][t][f])
            # print("*********************************** MM")




# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""

from __future__ import unicode_literals
from __future__ import print_function
from collections import defaultdict
from json import JSONEncoder
import threading
import json

metricsLock = threading.Lock()
requestCountLock = threading.Lock()

class Counter(object):
    def __init__(self):
        self.counter = 0
    def add_object(self):
        self.counter += 1
    def reset(self):
        self.counter = 0

class RouteCountMetric(dict):
    def __init__(self, *args, **kwargs):
        super(RouteCountMetric, self).__init__(*args, **kwargs)
        self["c"] = 0
        self["mx"] = 0
        self["mn"] = 0
        self["t"] = 0

class UserSessionTrackMetric(dict):
    def __init__(self, object_counter, user_id):
        dict.__init__(self)
        self.object_counter = object_counter
        self._track = defaultdict(list)
        self["uid"]=str(user_id)
        self["track"]=None
    def add_user_agent_ip_combo(self, user_agent, ip):
        if len(user_agent) > 256:
            user_agent = user_agent[:256]
        if (ip not in self._track[user_agent]):
            self._track[user_agent].append(ip)
            self["track"] = [[user_agent, self._track[user_agent]] for user_agent in self._track]
            self.object_counter.add_object()

class UserSessionMetric(list):
    def __init__(self, object_counter):
        self._tracks = dict()
        self.object_counter = object_counter
    def for_user_id(self, user_id):
        if user_id not in self._tracks:
            self.object_counter.add_object()
            self._tracks[user_id] = UserSessionTrackMetric(self.object_counter, user_id)
            self.append(self._tracks[user_id])
        return self._tracks[user_id]

class SessionsMetric(dict):
    def __init__(self,object_counter,*arg,**kw):
        dict.__init__(self,*arg,**kw)
        self.object_counter = object_counter
    def for_session(self, session_id):
        if session_id not in self:
            self[session_id] = UserSessionMetric(self.object_counter)
            self.object_counter.add_object()
        return self[session_id]

class AgentMetrics(object):
    def __init__(self):
        self.object_counter = Counter()
        self._session_metrics = SessionsMetric(self.object_counter)
        self._flush_signal = False

        self._route_count_table = defaultdict(RouteCountMetric)
        self._route_count_table_flush_signal = False

    def get_and_reset_sesssion_metric(self):
        session_metrics = None
        metricsLock.acquire()
        try:
            session_metrics = self._session_metrics
            self._session_metrics = SessionsMetric(self.object_counter)
            self._flush_signal = False
            self.object_counter.reset()
        finally:
            metricsLock.release()
        return session_metrics

    def get_object_count(self):
        return self.object_counter.counter

    def should_send_flush_event(self):
        metricsLock.acquire()
        try:
            if (self.object_counter.counter >= 250 and self._flush_signal == False):
                return True
        finally:
            metricsLock.release()
        return False

    def add_session_track_metric(self, session_id, user_id, user_agent, remote_addr):
        metricsLock.acquire()
        try:
            if (self.object_counter.counter >= 250):
                return
            self._session_metrics.for_session(session_id).for_user_id(user_id).add_user_agent_ip_combo(user_agent, remote_addr)
        except Exception as e:
            print(e)
        finally:
            metricsLock.release()

    def should_send_flush_event_for_route_count_table(self):
        requestCountLock.acquire()
        try:
            if (len(self._route_count_table) >= 150 and self._route_count_table_flush_signal == False):
                self._route_count_table_flush_signal = True
                return True
        finally:
            requestCountLock.release()
        return False

    def get_and_reset_route_count_table(self):
        route_count_table = None
        requestCountLock.acquire()
        try:
            route_count_table = dict(self._route_count_table)
            self._route_count_table = defaultdict(RouteCountMetric)
            self._route_count_table_flush_signal = False
        finally:
            requestCountLock.release()
        return route_count_table

    def add_route_count(self, route_id, request_time):
        requestCountLock.acquire()
        try:
            if (len(self._route_count_table) >= 150):
                return
            self._route_count_table[route_id]["c"] += 1
            self._route_count_table[route_id]["mx"] = max(self._route_count_table[route_id]["mx"], request_time)
            if self._route_count_table[route_id]["mn"] == 0:
                self._route_count_table[route_id]["mn"] = request_time
            else:
                self._route_count_table[route_id]["mn"] = min(self._route_count_table[route_id]["mn"], request_time)
            self._route_count_table[route_id]["t"] = ((self._route_count_table[route_id]["t"] * (self._route_count_table[route_id]["c"]-1))+request_time)/self._route_count_table[route_id]["c"]
        finally:
            requestCountLock.release()




import unittest
import re

from mock import patch
from tcell_agent.appsensor.rules.appsensor_rule_set import AppSensorRulePattern
from tcell_agent.appsensor.rules.appsensor_rule_set import AppSensorRuleSet

class AppSensorRuleSetTest(unittest.TestCase):

    def test_create_default(self):
        ruleset = AppSensorRuleSet()
        self.assertIsNone(ruleset.safe_pattern)
        self.assertEqual(ruleset.patterns, [])

    def test_add_pattern_from_dict_with_empty_dict(self):
        ruleset = AppSensorRuleSet()
        ruleset.add_pattern_from_dict({})
        self.assertIsNone(ruleset.safe_pattern)
        self.assertEqual(ruleset.patterns, [])

    def test_add_pattern_from_dict_with_common_pattern(self):
        ruleset = AppSensorRuleSet()
        ruleset.add_pattern_from_dict({"id": "1", "common": "common_regex"})
        self.assertIsNone(ruleset.safe_pattern)
        self.assertEqual(len(ruleset.patterns), 1)

        rule = ruleset.patterns[0]
        self.assertEqual(rule.pattern_id, "1")
        self.assertIsNotNone(rule.pattern_regex.search("common_regex ftw"))
        self.assertTrue(rule.enabled)

    def test_disable_common_pattern_by_overriding_pattern(self):
        ruleset = AppSensorRuleSet()
        ruleset.add_pattern_from_dict({"id": "1", "common": "common_regex", "python": "disabled"})
        self.assertIsNone(ruleset.safe_pattern)
        self.assertEqual(ruleset.patterns, [])

    def test_add_pattern_from_dict_with_python_pattern(self):
        ruleset = AppSensorRuleSet()
        ruleset.add_pattern_from_dict({"id": "1", "common": "common_regex", "python": "python_regex"})
        self.assertIsNone(ruleset.safe_pattern)
        self.assertEqual(len(ruleset.patterns), 1)

        rule = ruleset.patterns[0]
        self.assertEqual(rule.pattern_id, "1")
        self.assertIsNotNone(rule.pattern_regex.search("python_regex ftw"))
        self.assertTrue(rule.enabled)

    def check_violation_with_nil_param_value_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation(None, None, {}, True)
        self.assertIsNone(result)


    def check_violation_with_empty_param_value_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("", "", {}, True)
        self.assertIsNone(result)

    def check_violation_with_param_value_matching_safe_pattern_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "super_safe", {}, True)
        self.assertIsNone(result)

    def check_violation_with_param_value_matching_nothing_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "inane", {}, True)
        self.assertIsNone(result)

    def check_violation_with_param_value_matching_a_pattern_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "evil <script>", {}, True)
        self.assertIsNotNone(result)
        self.assertEqual(result["pattern"], "1")
        self.assertEqual(result["param"], "param_name")
        self.assertEqual(result["value"], "evil <script>")

    def check_violation_with_param_value_matching_a_pattern_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "evil <script>", {}, True)
        self.assertIsNotNone(result)
        self.assertEqual(result["pattern"], "1")
        self.assertEqual(result["param"], "param_name")
        self.assertEqual(result["value"], "evil <script>")

    def check_violation_with_param_value_matching_a_pattern_v1_compatability_off_no_patterns_enabled_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "evil <script>", {}, False)
        self.assertIsNone(result)

    def check_violation_with_param_value_matching_disabled_pattern_v1_compatability_off_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "evil <script>", {"2": True}, False)
        self.assertIsNone(result)

    def check_violation_with_param_value_matching_enabled_pattern_v1_compatability_off_test(self):
        ruleset = AppSensorRuleSet()
        ruleset.set_safe_pattern_from_string("super_safe")
        ruleset.add_pattern_from_dict({"id": "1", "common": "<(script)"})
        ruleset.add_pattern_from_dict({"id": "2", "common": "<(iframe)"})

        result = ruleset.check_violation("param_name", "evil <script>", {"1": True}, False)
        self.assertIsNotNone(result)
        self.assertEqual(result["pattern"], "1")
        self.assertEqual(result["param"], "param_name")
        self.assertEqual(result["value"], "evil <script>")



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

"""tCell Agent

This agent communicates with the tCell service and instruments your
application (when run as a command line tool) or can be used as an
SDK.

Example:
        $ tcell_agent python manage.py run 0.0.0.0:8000

https://www.tcell.io
"""

def init():
  from tcell_agent.instrumentation import run_instrumentations
  from tcell_agent.version import VERSION
  import traceback
  try:
    run_instrumentations()
    print("tCell.io Agent: [Info]  Started tCell.io agent v{version} (sitecustomize)".format(version=VERSION))
  except Exception as e:
    print("tCell.io Agent: [Error] Could not start tCell.io agent v{version} {e} (sitecustomize)".format(version=VERSION, e=e))
    print(e)
    traceback.print_exc()


from __future__ import unicode_literals
from __future__ import print_function

import json
import re
import unittest

from mock import Mock, ANY, patch

from tcell_agent.appsensor.meta import AppSensorMeta
from tcell_agent.appsensor.params import GET_PARAM
from tcell_agent.config import CONFIGURATION
from tcell_agent.policies.appsensor.payloads_policy import PayloadsPolicy


class PayloadsPolicyTest(unittest.TestCase):

    def false_log_payloads_log_test(self):
        policy_json = { "payloads": {"log_payloads": False} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = False

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)

    def true_log_payloads_false_log_appfirewall_events_log_test(self):
        policy_json = { "payloads": {"log_payloads": True} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = False

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)

    def true_log_payloads_true_log_appfirewall_events_no_blacklist_no_whitelist_log_test(self):
        policy_json = { "payloads": {"log_payloads": True} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)

    def true_log_payloads_true_log_appfirewall_events_no_blacklist_with_whitelist_missing_param_log_test(self):
        policy_json = { "payloads": {"log_payloads": True, "log_whitelist": {"username": ["*"]}} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)

    def true_log_payloads_true_log_appfirewall_events_no_blacklist_param_in_whitelist_but_incorrect_location_log_test(self):
        policy_json = {
            "payloads": {
                "log_payloads": True,
                "log_whitelist": {
                    "username": ["*"],
                    "password": ["cookie"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)

    def true_log_payloads_true_log_appfirewall_events_no_blacklist_param_in_whitelist_log_test(self):
        policy_json = {
            "payloads": {
                "log_payloads": True,
                "log_whitelist": {
                    "username": ["*"],
                    "password": ["form"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)

    def true_log_payloads_true_log_appfirewall_events_no_whitelist_with_blacklist_missing_param_log_test(self):
        policy_json = { "payloads": {"log_payloads": True, "log_blacklist": {"username": ["*"]}} }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        mock_logger = Mock()
        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger", return_value=mock_logger) as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            patched_get_payloadds_logger.assert_called_once_with()

            mock_logger.info.assert_called_once_with(ANY)
            call_args, call_kwargs = mock_logger.info.call_args
            self.assertEqual(json.loads(call_args[0]), {
                "event_type": "as",
                "remote_addr": "remote_addr",
                "pattern": "pattern",
                "m": "request_method",
                "param": "password",
                "meta": {"l": "query"},
                "sid": "session_id",
                "rid": "23947",
                "payload": "123password123",
                "dp": "xss",
                "uid": "user_id"
            })

    def true_log_payloads_true_log_appfirewall_events_no_whitelist_param_in_blacklist_but_incorrect_location_log_test(self):
        policy_json = {
            "payloads": {
                "log_payloads": True,
                "log_blacklist": {
                    "username": ["*"],
                    "password": ["cookie"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        mock_logger = Mock()
        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger", return_value=mock_logger) as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            patched_get_payloadds_logger.assert_called_once_with()

            mock_logger.info.assert_called_once_with(ANY)
            call_args, call_kwargs = mock_logger.info.call_args
            self.assertEqual(json.loads(call_args[0]), {
                "event_type": "as",
                "remote_addr": "remote_addr",
                "pattern": "pattern",
                "m": "request_method",
                "param": "password",
                "meta": {"l": "query"},
                "sid": "session_id",
                "rid": "23947",
                "payload": "123password123",
                "dp": "xss",
                "uid": "user_id"
            })

    def true_log_payloads_true_log_appfirewall_events_no_whitelist_param_in_blacklist_log_test(self):
        policy_json = {
            "payloads": {
                "log_payloads": True,
                "log_blacklist": {
                    "username": ["*"],
                    "password": ["form"]
                }
            }
        }
        policy = PayloadsPolicy()
        policy.from_json(policy_json)
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        old_allow = CONFIGURATION.log_appfirewall_events
        CONFIGURATION.log_appfirewall_events = True

        with patch("tcell_agent.policies.appsensor.payloads_policy.get_payloads_logger") as patched_get_payloadds_logger:
            policy.log("xss", meta, GET_PARAM, "password", "123password123", {"l": "query"}, "pattern")

            CONFIGURATION.log_appfirewall_events = old_allow

            self.assertFalse(patched_get_payloadds_logger.called)



# coding=utf-8

import unittest

from ...sanitize.sanitize_utils import SanitizeUtils
from ...config import CONFIGURATION


class SanitizerUtilsTest(unittest.TestCase):
    def old_header_test(self):
        CONFIGURATION.app_id = None
        response_info = {}
        response_info["headers"] = {"set-cookie":["sessionToken=abc123; Expires=Wed, 09 Jun 2021 10:18:14 GMT"]}
        SanitizeUtils.sanitize_response_info(response_info)
        self.assertEqual(response_info["headers"]["set-cookie"], ['sessionToken=8aed87a14d7369b35600b3eb61d24e51dab0fe083749b270acb9be31603da8d6; expires=Wed, 09 Jun 2021 10:18:14 GMT'])

    def request_sanitize_test(self):
        CONFIGURATION.app_id = None
        request_info = {}
        request_info["uri"] = "/abc/def"
        request_info["headers"] = {"cookie":["sessionToken=abc123; sessionToken2=sdfdsf"]}
        SanitizeUtils.sanitize_request_info(request_info)
        self.assertEqual(
            request_info["headers"]["cookie"],
            [' sessionToken=8aed87a14d7369b35600b3eb61d24e51dab0fe083749b270acb9be31603da8d6; sessionToken2=6255969f6b578c815cc57426ad8520f5480eff23bb9f632720a38d2b2c9429aa'])

    def strip_uri_test(self):
        CONFIGURATION.app_id = None
        self.assertEqual(SanitizeUtils.strip_uri("/abc/def?bbb=abcd"), "/abc/def?bbb=")
        self.assertEqual(SanitizeUtils.strip_uri("https://aaa:8000/abc/def?bbb=ccbc"), "https://aaa:8000/abc/def?bbb=")

    def strip_query_string_test(self):
        self.assertEqual(SanitizeUtils.strip_uri("/abc/def?%C2%B5=%C3%A9%3Cscript%3E%C2%B5"), "/abc/def?%C2%B5=")
        self.assertEqual(
            SanitizeUtils.strip_uri("http://192.168.99.100:3000/account/signup/?%C2%B5=%C3%A9%3Cscript%3E%C2%B5"),
            "http://192.168.99.100:3000/account/signup/?%C2%B5=")

    def strip_path_string_test(self):
        CONFIGURATION.app_id = None
        self.assertEqual(
            SanitizeUtils.strip_uri("/vulnerabilities/%C2%B5%3Cscript%3E/?bbb=ccbc"),
            "/vulnerabilities/%C2%B5%3Cscript%3E/?bbb=")
        self.assertEqual(
            SanitizeUtils.strip_uri("http://192.168.99.100:3000/vulnerabilities/%C2%B5%3Cscript%3E/?bbb=ccbc"),
            "http://192.168.99.100:3000/vulnerabilities/%C2%B5%3Cscript%3E/?bbb=")



from __future__ import unicode_literals
from __future__ import print_function

import unittest
from ...agent import TCellAgent

class AppSensorPolicyTest(unittest.TestCase):
    def simple_database_route_test(self):
        events = []
        def addEvent(self, event):
          event.post_process()
          events.append(event)

        TCellAgent.addEvent = addEvent

        TCellAgent.discover_database_fields("databasex","schemax","tablex",["fieldx"],"routex")
        self.assertEqual(len(events), 0)

        TCellAgent.tCell_agent = TCellAgent()
        policy_json = {
            "dlp":{
              "policy_id":"nyzd", 
              "data":{
                  "data_discovery": {
                    "database_enabled": True
                  },
              }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        TCellAgent.discover_database_fields("databasex","schemax","tablex",["fieldx"],"routex")
        TCellAgent.discover_database_fields("databasex","schemax","tablex",["fieldx"],"routex")
        TCellAgent.discover_database_fields("databasex","schemax","tablex",["fieldx"],"routex")
        TCellAgent.discover_database_fields("databasex","schemax","tablex",["fieldx"],"routex")

        self.assertEqual(TCellAgent.tCell_agent.route_table.routes["routex"].database_fields[hash("databasex,schemax,tablex," + ",".join(["fieldx"]))].discovered, True)
        self.assertEqual(TCellAgent.tCell_agent.route_table.routes["routex"].database_fields[hash("databasex,schemax,tabley," + ",".join(["fieldx"]))].discovered, False)

        self.assertEqual(len(events), 1)


# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved
import re

from django.http import HttpResponse

from tcell_agent.agent import TCellAgent, PolicyTypes
from tcell_agent.config import CONFIGURATION
from tcell_agent.instrumentation import safe_wrap_function


class BodyFilterMiddleware(object):

    def add_tag(self, request, response):
        if isinstance(response, HttpResponse) and response.has_header("Content-Type"):
            if response["Content-Type"] and response["Content-Type"].startswith("text/html"):
                response_type = type(response.content)
                tag_policy = TCellAgent.get_policy(PolicyTypes.CSP)
                if tag_policy and tag_policy.js_agent_api_key:
                    if CONFIGURATION.js_agent_api_base_url:
                        replace = "<head><script src=\"" + CONFIGURATION.js_agent_url + "\" tcellappid=\"" + CONFIGURATION.app_id + "\" tcellapikey=\"" + tag_policy.js_agent_api_key + "\" tcellbaseurl=\""+ CONFIGURATION.js_agent_api_base_url +"\"></script>"
                    else:
                        replace = "<head><script src=\"" + CONFIGURATION.js_agent_url + "\" tcellappid=\"" + CONFIGURATION.app_id + "\" tcellapikey=\"" + tag_policy.js_agent_api_key + "\"></script>"

                    if isinstance(response.content, str) is False:
                        replace = replace.encode("utf-8")

                    try:
                        if response_type == str:
                            response.content = re.sub(b"<head>", replace, response.content.decode('utf8'))

                        else:
                            response.content = re.sub(b"<head>", replace, response.content)
                    except UnicodeDecodeError:
                        pass


    def process_request(self, request):
        pass

    def process_response(self, request, response):
        safe_wrap_function("Insert Body Tag", self.add_tag, request, response)
        return response



from __future__ import unicode_literals
from __future__ import print_function

import unittest
from ...policies.login_policy import LoginPolicy

import json

policy_one = """
{
  "policy_id":"00a1",
  "data": {
    "options": {
       "login_failed_enabled":true,
       "login_success_enabled":true
    }
  }
}
"""

class LoginPolicyTest(unittest.TestCase):
    def classname_test(self):
        self.assertEqual(LoginPolicy.api_identifier, "login")

    def read_appensor_policy_test(self):
        policy_json = json.loads(policy_one)
        policy = LoginPolicy()
        self.assertEqual(policy.login_failed_enabled, False)
        self.assertEqual(policy.login_success_enabled, False)
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.login_failed_enabled, True)
        self.assertEqual(policy.login_success_enabled, True)




# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function


from tcell_agent.sanitize import SanitizeUtils

from . import SensorEvent

class AppSensorEvent(SensorEvent):
    def __init__(self,
                 detection_point,
                 parameter,
                 location,
                 remote_address,
                 route_id,
                 meta,
                 method,
                 hmacd_session_id=None,
                 user_id=None,
                 count=None,
                 payload=None,
                 pattern=None):
        super(AppSensorEvent, self).__init__("as")
        self["dp"] = detection_point
        self["param"] = parameter
        self._raw_location = location
        self["remote_addr"] = remote_address
        self["m"] = method
        if meta is not None:
            self["meta"] = meta
        if route_id is not None:
            self["rid"] = str(route_id)
        if user_id is not None:
            self["uid"] = str(user_id)
        if hmacd_session_id is not None:
            self["sid"] = hmacd_session_id
        if count is not None:
            self["count"] = count
        if payload is not None:
            self["payload"] = payload[:150]
        if pattern is not None:
            self["pattern"] = pattern

    def post_process(self):
        if "payload" in self:
            self["payload"] = self["payload"][:150]
        if self._raw_location is not None:
            self["uri"] = SanitizeUtils.strip_uri(self._raw_location)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

import traceback

from .manager import InstrumentationManager

_test_mode = False

def run_instrumentations():
    """Call the instrumentation functions, only works if they exist"""
    if (InstrumentationManager.initial_instrumentation_run == False):
        InstrumentationManager.initial_instrumentation_run = True
        from tcell_agent.instrumentation import gunicorn_tcell
        from tcell_agent.instrumentation import django
        from tcell_agent.instrumentation import flask

def safe_wrap_function(description, func, *args, **kwargs):
  if (_test_mode):
    print("[tcell] >" + description + "<")
    return func(*args, **kwargs)
  try:
    return func(*args, **kwargs)
  except Exception as e:
    handle_exception(description, e)

def safe_wrap_function_no_log(description, func, *args, **kwargs):
  if (_test_mode):
    print("[tcell] >" + description + "<")
    return func(*args, **kwargs)
  try:
    return func(*args, **kwargs)
  except Exception as e:
    pass

def handle_exception(msg, e, logger=None):
  if logger:
    logger.error(e)
  else:
    print(msg)
    traceback.print_exc()

class BaseWrapper(object):
    def __init__(self, instance):
        self._instance = instance

    def __getattr__(self, attr):
        #print("Calling __getattr__: "+attr)
        if hasattr(self._instance, attr):
            def wrapper(*args, **kw):
                #print('called with %r and %r' % (args, kw))
                return getattr(self._instance, attr)(*args, **kw)
            return wrapper
        raise AttributeError(attr)


# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from future.utils import iteritems
from tcell_agent.appsensor import params
from tcell_agent.appsensor.params import GET_PARAM, POST_PARAM, JSON_PARAM, COOKIE_PARAM, URI_PARAM
from tcell_agent.config import CONFIGURATION
from tcell_agent.instrumentation import safe_wrap_function
from tcell_agent.policies import TCellPolicy
from tcell_agent.policies.appsensor import CmdiSensor
from tcell_agent.policies.appsensor import DatabaseSensor
from tcell_agent.policies.appsensor import FptSensor
from tcell_agent.policies.appsensor import LoginSensor
from tcell_agent.policies.appsensor import MiscSensor
from tcell_agent.policies.appsensor import NullbyteSensor
from tcell_agent.policies.appsensor import RequestSizeSensor
from tcell_agent.policies.appsensor import ResponseCodesSensor
from tcell_agent.policies.appsensor import ResponseSizeSensor
from tcell_agent.policies.appsensor import RetrSensor
from tcell_agent.policies.appsensor import SqliSensor
from tcell_agent.policies.appsensor import UserAgentSensor
from tcell_agent.policies.appsensor import XssSensor
from tcell_agent.policies.appsensor import PayloadsPolicy

class AppSensorPolicy(TCellPolicy):
    api_identifier = "appsensor"

    options = [
        "req_res_size",
        "resp_codes",
        "xss",
        "sqli",
        "cmdi",
        "fpt",
        "null",
        "retr",
        "login_failure",
        "ua",
        "errors",
        "database"]

    options_v2_classes = {
        "req_size": RequestSizeSensor,
        "resp_size": ResponseSizeSensor,
        "resp_codes": ResponseCodesSensor,
        "xss": XssSensor,
        "sqli": SqliSensor,
        "cmdi": CmdiSensor,
        "fpt": FptSensor,
        "nullbyte": NullbyteSensor,
        "retr": RetrSensor,
        "login": LoginSensor,
        "ua": UserAgentSensor,
        "errors": MiscSensor,
        "database": DatabaseSensor}

    def __init__(self, policy_json=None):
        super(AppSensorPolicy, self).__init__()
        self.init_variables()
        if policy_json is not None:
            self.loadFromJson(policy_json)

    def init_variables(self):
        self.enabled = False
        self.payloads_policy = PayloadsPolicy()
        self.options = {}

    def run_for_request(self, appsensor_meta, json_body):
        self.check_request_size(appsensor_meta)

        self.check_params_for_injections(appsensor_meta, json_body)

        if "ua" in self.options:
            safe_wrap_function(
                "Check User Agent",
                self.options["ua"].check,
                appsensor_meta
            )


    def run_for_response(self, appsensor_meta):
        self.check_response_size(appsensor_meta)
        self.check_response_code(appsensor_meta)

    def run_for_path_parameters(self, appsensor_meta):
        path_dict = appsensor_meta.path_dict

        for param_name, param_value in iteritems(path_dict or {}):
            safe_wrap_function(
                "Check PATH var injections",
                self.check_param_for_injections,
                URI_PARAM,
                appsensor_meta,
                param_name,
                param_value
            )


    def check_login_failure(self, appsensor_meta, username):
        if "login" in self.options:
            safe_wrap_function(
                "Check Login Failure",
                self.options["login"].check,
                appsensor_meta,
                username
            )

    def check_db_rows(self, appsensor_meta, number_of_records):
        if "database" in self.options:
            safe_wrap_function(
                "Appsensor Check Number of DB Rows",
                self.options["database"].check,
                appsensor_meta,
                number_of_records
            )

    def check_param_for_injections(self, param_type, appsensor_meta, param_name, param_value):
        if self.options.get("xss") and self.options["xss"].check(param_type, appsensor_meta, param_name, param_value, self.payloads_policy):
            return
        if self.options.get("sqli") and self.options["sqli"].check(param_type, appsensor_meta, param_name, param_value, self.payloads_policy):
            return
        if COOKIE_PARAM != param_type:
            if self.options.get("cmdi") and self.options["cmdi"].check(param_type, appsensor_meta, param_name, param_value, self.payloads_policy):
                return
        if COOKIE_PARAM != param_type:
            if self.options.get("fpt") and self.options["fpt"].check(param_type, appsensor_meta, param_name, param_value, self.payloads_policy):
                return
        if COOKIE_PARAM != param_type:
            if self.options.get("nullbyte") and self.options["nullbyte"].check(param_type, appsensor_meta, param_name, param_value, self.payloads_policy):
                return
        if GET_PARAM == param_type or COOKIE_PARAM == param_type:
            if self.options.get("retr") and self.options["retr"].check(param_type, appsensor_meta, param_name, param_value, self.payloads_policy):
                return

    def check_params_for_injections(self, appsensor_meta, json_body):
        get_dict = params.flatten_clean(appsensor_meta.get_dict)
        cookie_dict = params.flatten_clean(appsensor_meta.cookie_dict)
        json_body = params.flatten_clean(json_body)

        post_dict = appsensor_meta.post_dict
        files_dict = appsensor_meta.files_dict

        for param_name, param_value in iteritems(get_dict):
            param_name = param_name[-1]
            safe_wrap_function(
                "Check GET var injections",
                self.check_param_for_injections,
                GET_PARAM,
                appsensor_meta,
                param_name,
                param_value
            )

        for param_name, param_value in iteritems(post_dict):
            param_name = param_name[-1]
            safe_wrap_function(
                "Check POST var injections",
                self.check_param_for_injections,
                POST_PARAM,
                appsensor_meta,
                param_name,
                param_value
            )

        for param_name, param_value in iteritems(files_dict):
            param_name = param_name[-1]
            safe_wrap_function(
                "Check Filename injections",
                self.check_param_for_injections,
                POST_PARAM,
                appsensor_meta,
                param_name,
                param_value
            )

        for param_name, param_value in iteritems(cookie_dict or {}):
            param_name = param_name[-1]
            safe_wrap_function(
                "Check Cookies var injections",
                self.check_param_for_injections,
                COOKIE_PARAM,
                appsensor_meta,
                param_name,
                param_value
            )

        for param_name, param_value in iteritems(json_body or {}):
            param_name = param_name[-1]
            safe_wrap_function(
                "Check JSON var injections",
                self.check_param_for_injections,
                JSON_PARAM,
                appsensor_meta,
                param_name,
                param_value
            )


    def check_request_size(self, appsensor_meta):
        if "req_size" in self.options:
            safe_wrap_function(
                "Check Request Size",
                self.options["req_size"].check,
                appsensor_meta,
                appsensor_meta.request_content_len)

    def check_response_size(self, appsensor_meta):
        if "resp_size" in self.options:
            safe_wrap_function(
                "Check Response Size",
                self.options["resp_size"].check,
                appsensor_meta,
                appsensor_meta.response_content_len)

    def check_response_code(self, appsensor_meta):
        if "resp_codes" in self.options:
            safe_wrap_function(
                "Check Response Codes",
                self.options["resp_codes"].check,
                appsensor_meta,
                appsensor_meta.response_code)

    def csrf_rejected(self, appsensor_meta, reason):
        if "errors" in self.options:
            safe_wrap_function(
                "CSRF Exception processing",
                self.options["errors"].csrf_rejected,
                appsensor_meta,
                reason)

    def sql_exception_detected(self, database, appsensor_meta, exc_type, exc_value, traceback):
        if "errors" in self.options:
            safe_wrap_function(
                "SQL Exception processing",
                self.options["errors"].sql_exception_detected,
                database,
                appsensor_meta,
                exc_type,
                exc_value,
                traceback)

    def loadFromJson(self, policy_json):
        if "policy_id" in policy_json:
            self.policy_id = policy_json["policy_id"]
        else:
            raise Exception("Policy Id Not Found")

        self.init_variables()

        policy_data = policy_json.get("data")

        if "version" in policy_json and policy_json["version"] == 2:
            if policy_data:
                sensors_json = policy_data.get("sensors")
                if sensors_json:
                    self.payloads_policy.from_json(policy_data.get("options", {}))

                    for option, clazz in iteritems(self.options_v2_classes):
                        settings = sensors_json.get(option, {})
                        settings["enabled"] = option in sensors_json
                        self.options[option] = clazz(settings)

        else:
            if policy_data:
                options_json = policy_data.get("options")
                if options_json:
                    self.payloads_policy.from_json({
                        "payloads": {
                            "send_payloads": True,
                            "log_payloads": True
                        }
                    })

                    for option in AppSensorPolicy.options:
                        if "req_res_size" == option:
                            enabled = options_json.get(option, False)
                            self.options["req_size"] = RequestSizeSensor({"enabled": enabled})
                            self.options["resp_size"] = ResponseSizeSensor({"enabled": enabled})
                        elif "resp_codes" == option:
                            enabled = options_json.get(option, False)
                            self.options[option] = ResponseCodesSensor({
                                "enabled": enabled,
                                "series_400_enabled": True,
                                "series_500_enabled": True})
                        elif "null"  == option:
                            enabled = options_json.get(option, False)
                            self.options["nullbyte"] = NullbyteSensor({"enabled": enabled, "v1_compatability_enabled": True})
                        elif "login_failure"  == option:
                            enabled = options_json.get(option, False)
                            self.options["login"] = LoginSensor({"enabled": enabled})
                        elif "ua" == option:
                            self.options[option] = UserAgentSensor({"enabled": False, "empty_enabled": False})
                        elif "errors" == option:
                            self.options[option] = MiscSensor({"csrf_exception_enabled": False, "sql_exception_enabled": False})
                        else:
                            enabled = options_json.get(option, False)
                            clazz = self.options_v2_classes[option]
                            self.options[option] = clazz({"enabled": enabled, "v1_compatability_enabled": True})



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.agent import TCellAgent, PolicyTypes
from tcell_agent.appsensor.django import appsensor_login_failed

import os
from tcell_agent.sensor_events import LoginEvent, HoneytokenSensorEvent
from tcell_agent.instrumentation import handle_exception, safe_wrap_function
from tcell_agent.instrumentation.django.middleware.globalrequestmiddleware import GlobalRequestMiddleware

from tcell_agent.instrumentation.django.utils import header_keys_from_django_request

import logging
import tcell_agent.tcell_logger
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

def _instrument_common_auth():
    pass
    # old_process_authenticate = AuthenticationForm.clean
    # def clean(self):
    #     try:
    #         retval = old_process_authenticate(self)
    #             if (self.user_cache is not None):
    #                 print("User logged in")
    #                 self.
    #     except Exception as e:
    #         raise e
    #     return retval
    # AuthenticationForm.clean = clean

def _userLoggedIn(sender, user, request, **kwargs):
    def _set_login_success():
        #request._tcell_signals['login_success'] = True
        # put appsensor stuff here
        pass
    def _set_loginfraud_success():
        login_policy = TCellAgent.get_policy(PolicyTypes.LOGIN)
        if login_policy is None or login_policy.login_success_enabled != True:
            return
        request = GlobalRequestMiddleware.get_current_request()
        if request is not None:
            username = None
            try:
                username = user.get_username()
            except:
                pass
            if request:
                try:
                    if request.method == "POST":
                        username = request.POST.get("user", username)
                        username = request.POST.get("email", username)
                        username = request.POST.get("email_address", username)
                        username = request.POST.get("username", username)
                    else:
                        username = request.GET.get("username", username)
                except Exception as e:
                    LOGGER.error("Could not determine username for login success event")
                    LOGGER.debug(e)
            LOGGER.debug("Login success event for {username}".format(username=username))
            event = LoginEvent().success(
                user_id=username, 
                user_agent=request.META.get("HTTP_USER_AGENT", None),
                referrer=request.META.get("HTTP_REFERER", None),
                remote_addr=request._tcell_context.remote_addr,
                header_keys=header_keys_from_django_request(request),
                document_uri=request._tcell_context.fullpath,
                session_id=request._tcell_context.session_id)
            TCellAgent.send(event)
    safe_wrap_function("Setting login success", _set_login_success) 
    safe_wrap_function("LoginFraud login success", _set_loginfraud_success) 

def _userLoginFailed(sender, credentials, **kwargs):
    def _set_loginfraud_failure():
        login_policy = TCellAgent.get_policy(PolicyTypes.LOGIN)
        if login_policy is None or login_policy.login_failed_enabled != True:
            return
        request = GlobalRequestMiddleware.get_current_request()
        if request is not None:
            username = None
            if credentials:
                username = credentials.get("username")
            LOGGER.debug("Login failed event for {username}".format(username=username))
            event = LoginEvent().failure(
                user_id=username, 
                user_agent=request.META.get("HTTP_USER_AGENT", None),
                referrer=request.META.get("HTTP_REFERER", None),
                remote_addr=request._tcell_context.remote_addr,
                header_keys=header_keys_from_django_request(request),
                document_uri=request._tcell_context.fullpath,
                session_id=request._tcell_context.session_id)
            TCellAgent.send(event)
    safe_wrap_function("LoginFraud login failure", _set_loginfraud_failure) 

def _addUserLoginSignals():
    user_logged_in.connect(_userLoggedIn)
    user_login_failed.connect(_userLoginFailed)

def _addAuthIntercept():
    original_auth = ModelBackend.authenticate
    def authenticate(self, username=None, password=None, **kwargs):
        def _check_honeytoken_creds():
            request = GlobalRequestMiddleware.get_current_request()
            if request:
                honeytoken_policy = TCellAgent.get_policy(PolicyTypes.HONEYTOKEN)
                if honeytoken_policy:
                    token_id = honeytoken_policy.get_id_for_credential(username, password)
                    if token_id:
                        event = HoneytokenSensorEvent(
                            token_id=token_id,
                            remote_addr=request._tcell_context.remote_addr
                        )
                        TCellAgent.send(event)
        safe_wrap_function("Honeytokens check", _check_honeytoken_creds) 
        return original_auth(self, username, password, **kwargs)
    ModelBackend.authenticate = authenticate

try:
    import django
    from django.contrib.auth.forms import AuthenticationForm
    if TCellAgent.tCell_agent:
        _instrument_common_auth()
        from django.contrib.auth.signals import user_logged_in, user_login_failed
        from django.db.backends.signals import connection_created
        safe_wrap_function("Adding user login signals", _addUserLoginSignals)
        from django.contrib.auth.backends import ModelBackend
        safe_wrap_function("Adding user honeytoken code", _addAuthIntercept)
except Exception as e:
    LOGGER.debug("Could not instrument django common-auth")
    LOGGER.debug(e)




# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

import imp
import os
import sys
import future

tcell_pythonpath = os.path.dirname(__file__)
path = list(sys.path)
if tcell_pythonpath in path:
    del path[path.index(tcell_pythonpath)]

def find_dotted_module(name, path=None):
    '''
    Example: find_dotted_module('mypackage.myfile')

    Background: imp.find_module() does not handle hierarchical module names (names containing dots).

    Important: No code of the module gets executed. The module does not get loaded (on purpose)
    ImportError gets raised if the module can't be found.

    Use case: Test discovery without loading (otherwise coverage does not see the lines which are executed 
              at import time)
    '''

    for x in name.split('.'):
        if path is not None:
            path=[path]
        file, path, descr = imp.find_module(x, path)
    return file, path, descr

import traceback
for p in path:
    try:
        if "newrelic" not in p:
            (file, pathname, description) = find_dotted_module("sitecustomize", p)

            imp.load_module('sitecustomize', file, pathname, description)
            break
    except Exception as e:
        pass

from tcell_agent.instrumentation import run_instrumentations
from tcell_agent.version import VERSION
try:
     run_instrumentations()
     print("tCell.io Agent: [Info]  Started tCell.io agent v{version} (sitecustomize)".format(version=VERSION))
except Exception as e:
    print("tCell.io Agent: [Error] Could not start tCell.io agent v{version} {e} (sitecustomize)".format(version=VERSION, e=e))
    traceback.print_exc()



from .appsensor_rule_manager import AppSensorRuleManager

import os.path
import sys
import json

import logging
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

rule_manager = AppSensorRuleManager()
try:
    basepath = os.path.dirname(__file__)
    baserulesfilename = os.path.abspath(os.path.join(basepath, "python-baserules.json"))
    rule_manager.load_rules_file(baserulesfilename)
except Exception as read_rules_exception:
    print(read_rules_exception)
    logging.debug(read_rules_exception)


from tcell_agent.agent.agent_threads import TCellAgent
def just_return(self):
    return
TCellAgent.ensure_polling_thread_running = just_return
TCellAgent.ensure_metrics_pipe_thread_running = just_return
TCellAgent.ensure_fork_pipe_thread_running = just_return
TCellAgent.ensure_event_handler_thread_running = just_return
from tcell_agent.appsensor.manager import app_sensor_manager
app_sensor_manager.use_threads = False


from tcell_agent.config import configuration
configuration.WARN_TO_CONSOLE = False